#!/usr/bin/env python3

from collections import Counter
import io
from typing import Iterable, Sequence

import matplotlib.pyplot as plt

from turnips.meta import MetaModel
from turnips.model import TRIPLE, SPIKE, DECAY, BUMP, Model, ModelEnum
from turnips.multi import MultiModel
from turnips.ttime import TimePeriod
from turnips.markov import MARKOV


def plot_models_range(name: str,
                      models: Sequence[Model],
                      previous: ModelEnum,
                      add_points: bool = False) -> None:
    '''
    Plot a fill_between for all models' low and high values using an
    alpha (transparency) equal to 1/num_models. Plot a regular line
    for all fixed prices.

    Shows ~probability of various prices based on your possible models.
    '''

    colors = {
        TRIPLE: 'orange',
        SPIKE:  'green',
        DECAY:  'red',
        BUMP:   'blue',
    }

    _fig, ax = plt.subplots()

    # cosmetics
    ax.set_title(f'Island {name}: current: !!ERROR!!; Last: {previous.name}')
    ax.set_ylabel('Turnip Price')
    ax.set_xticklabels(['Mon AM', 'Mon PM', 'Tue AM', 'Tue PM', 'Wed AM', 'Wed PM',
                        'Thu AM', 'Thu PM', 'Fri AM', 'Fri PM', 'Sat AM', 'Sat PM'])
    ax.xaxis.set_ticks(range(2, 14))
    plt.xticks(rotation=45)
    plt.grid(axis='both', which='both', ls='--')
    ax.set_ylim(0, 660)
    plt.tight_layout()

    if len(models) == 0:
        return

    a_model = models[0]

    continuous_priced_days = []
    continuous_priced_chunk = set()
    continuous_unpriced_days = []
    continuous_unpriced_chunk = set()
    for day in range(2, 14):
        # does this day have data?
        if a_model.timeline[TimePeriod(day)].price.is_atomic:
            # is tomorrow a valid day to have data?
            if day < 13:
                # does tomorrow have data?
                if a_model.timeline[TimePeriod(day + 1)].price.is_atomic:
                    # build onto the chunk
                    continuous_priced_chunk.update([day, day + 1])
                # chunk broken.
                else:
                    continuous_priced_chunk.update([day])
                    continuous_priced_days.append(list(continuous_priced_chunk))
                    continuous_priced_chunk = set()
            else:
                # end of the week, finish the priced_days
                continuous_priced_days.append(list(continuous_priced_chunk))
        # today does not have data
        else:
            # is tomorrow a valid day to have data?
            if day < 13:
                # does it?
                if not a_model.timeline[TimePeriod(day + 1)].price.is_atomic:
                    # build the chunk
                    if day != 2:
                        # add yesterday unless today is monday_am
                        continuous_unpriced_chunk.update([day - 1, day, day + 1])
                    else:
                        continuous_unpriced_chunk.update([day, day + 1])
                # chunk broken
                else:
                    if day != 2:
                        continuous_unpriced_chunk.update([day - 1, day, day + 1])
                    else:
                        continuous_unpriced_chunk.update([day, day + 1])
                    continuous_unpriced_days.append(list(sorted(continuous_unpriced_chunk, key=lambda x: x)))
                    continuous_unpriced_chunk = set()
            else:
                # end of the week, finish the unpriced_days
                continuous_unpriced_days.append(list(continuous_unpriced_chunk))

    for chunk in continuous_priced_days:
        vals = [a_model.timeline[TimePeriod(day)].price.value for day in chunk]
        plt.plot(chunk, vals, c='black')

    model_counts = Counter(x.model_type for x in models)
    remaining_model_types = model_counts.keys()
    remaining_probability = sum(MARKOV[previous][rem_mod]
                                for rem_mod in remaining_model_types)
    adjusted_priors = {model: MARKOV[previous][model] / remaining_probability
                       for model in model_counts.keys()}

    for chunk in continuous_unpriced_days:
        if len(chunk) == 1:
            # if this is one day of unpriced data, connect it to the neighbors.
            value = chunk[0]
            chunk = [value - 1, value, value + 1]
        for model in models:
            low_vals = [model.timeline[TimePeriod(day)].price.lower for day in chunk]
            high_vals = [model.timeline[TimePeriod(day)].price.upper for day in chunk]

            if previous != ModelEnum.unknown:
                alpha = adjusted_priors[model.model_type] / model_counts[model.model_type]
            else:
                alpha = 1 / len(models)

            plt.fill_between(chunk, low_vals, high_vals, alpha=alpha, color=colors[model.model_type])

            if add_points:
                plt.scatter(chunk, low_vals, c='black', s=2)
                plt.scatter(chunk, high_vals, c='black', s=2)

    # cosmetics
    msummary = '+'.join(['{}_{{{}}}^{{{:.2f}}}'.format(t, l.name, adjusted_priors[l])
                         for l, t in model_counts.items()])
    ax.set_title(f'Island {name}: ${len(models)}_{{total}}={msummary}$, Last: {previous.name}')


def plot_models_range_interactive(name: str,
                                  models: Sequence[Model],
                                  previous: ModelEnum,
                                  add_points: bool = False) -> None:
    '''
    Plot a fill_between for all models' low and high values using an
    alpha (transparency) equal to 1/num_models. Plot a regular line
    for all fixed prices.

    Shows ~probability of various prices based on your possible models.
    '''
    plot_models_range(name, models, previous, add_points)
    plt.show()


def plot_models_range_data(name: str,
                           models: Sequence[Model],
                           previous: ModelEnum,
                           add_points: bool = False) -> None:
    '''
    Plot a fill_between for all models' low and high values using an
    alpha (transparency) equal to 1/num_models. Plot a regular line
    for all fixed prices.

    Shows ~probability of various prices based on your possible models.
    '''
    plot_models_range(name, models, previous, add_points)
    buff = io.BytesIO()
    plt.savefig(buff)
    return buff.getvalue()


def global_plot(island_models: Iterable[MultiModel]) -> None:
    '''Plot a histogram, per day, for all possible prices in the archipelago.

    This isn't exactly a probability curve since the histogram isn't scaled
    by the real probabilities.'''

    arch = MetaModel(100, island_models)

    hist = arch.histogram()

    num_ticks = len(range(2, 14))
    _fig, ax = plt.subplots(num_ticks, sharex=True)

    for i, (time, pricecounts) in enumerate(hist.items()):
        ax[i].hist(list(pricecounts.elements()), 50)
        ax[i].text(.8, .5, time,
                   horizontalalignment='center',
                   transform=ax[i].transAxes)

    ax[i].set_xlabel('Turnip Price')

    plt.show()



#!/usr/bin/env python3

from __future__ import annotations

from collections import Counter
import enum
import math
import random
from typing import Dict, Generator, List, Optional, Type
from typing import Counter as CounterType

from turnips.price import Price
from turnips.modifier import (
    Modifier,
    SmallProfit,
    MediumProfit,
    LargeProfit,
    MediumLoss,
    WideLoss,
    RapidDecay,
    InitialDecay,
    SlowDecay,
    CappedPassthrough,
    Passthrough,
)
from turnips.ttime import AnyTime, TimePeriod

class ModelEnum(enum.Enum):
    unknown = -1
    triple = 0
    spike = 1
    decay = 2
    bump = 3

UNKNOWN = ModelEnum.unknown
TRIPLE = ModelEnum.triple
SPIKE = ModelEnum.spike
DECAY = ModelEnum.decay
BUMP = ModelEnum.bump


class Model:
    _model_class: ModelEnum

    def __init__(self, initial: Price):
        self.initial = initial
        self.timeline: Dict[TimePeriod, Modifier] = {}

    @property
    def model_type(self) -> ModelEnum:
        return self._model_class

    @property
    def model_name(self) -> str:
        return self.model_type.name

    @property
    def params(self) -> str:
        raise NotImplementedError("Abstract Model has no Parameters")

    @property
    def name(self) -> str:
        basis = [f"{self.model_name}@{str(self.initial)}"]
        if self.params:
            basis.append(self.params)
        return "; ".join(basis)

    def __str__(self) -> str:
        lines = [
            f"# Model: {self.name}\n",
        ]
        for time, modifier in self.timeline.items():
            col = f"{time.name}:"
            price = f"{str(modifier.price)}"
            modrange = f"{str(modifier)}"
            modtype = f"{modifier.name}"
            lines.append(f"{col:13} {price:10} {modrange:20} ({modtype})")
        return '\n'.join(lines)

    def histogram(self) -> Dict[str, CounterType[int]]:
        histogram: Dict[str, CounterType[int]] = {}
        for time, modifier in self.timeline.items():
            price_low = math.ceil(modifier.price.lower)
            price_high = math.ceil(modifier.price.upper)
            counter = histogram.setdefault(time.name, Counter())
            for price in range(price_low, price_high + 1):
                counter[price] += 1
        return histogram

    @classmethod
    def inner_permutations(cls, initial: int) -> Generator[Model, None, None]:
        yield cls(Price(initial))

    @classmethod
    def permutations(cls, initial: Optional[int] = None) -> Generator[Model, None, None]:
        if initial is not None:
            assert 90 <= initial <= 110, "Initial should be within [90, 110]"
            lower = initial
            upper = initial
        else:
            lower = 90
            upper = 110

        for tprice in range(lower, upper + 1):
            yield from cls.inner_permutations(tprice)

    def fix_price(self, time: AnyTime, price: int) -> None:
        timeslice = TimePeriod.normalize(time)
        self.timeline[timeslice].fix_price(price)

    def random_price(self, time: AnyTime):
        '''Provide a random price for the given time, unless that time has data.'''
        timeslice = TimePeriod.normalize(time)
        if self.timeline[timeslice].price.is_atomic:
            return self.timeline[timeslice].price.value
        return random.randint(self.timeline[timeslice].price.lower,
                              self.timeline[timeslice].price.upper)


class TripleModel(Model):
    _model_class = ModelEnum.triple

    def __init__(self,
                 initial: Price,
                 length_phase1: int,
                 length_decay1: int,
                 length_phase2: int):

        super().__init__(initial)

        if not 0 <= length_phase1 <= 6:
            raise ValueError("Phase1 length must be between [0, 6]")
        self._length_phase1 = length_phase1

        if not 2 <= length_decay1 <= 3:
            raise ValueError("Decay1 length must be 2 or 3")
        self._length_decay1 = length_decay1
        self._length_decay2 = 5 - length_decay1

        remainder = 7 - length_phase1
        if not 1 <= length_phase2 <= remainder:
            raise ValueError(f"Phase2 must be between [1, {remainder}]")
        self._length_phase2 = length_phase2
        self._length_phase3 = remainder - length_phase2

        assert (self._length_phase1
                + self._length_phase2
                + self._length_phase3
                + self._length_decay1
                + self._length_decay2) == 12

        chain: List[Modifier] = []
        decay_class: Type[Modifier]

        def _push_node(mod_cls: Type[Modifier]) -> None:
            mod = mod_cls(self.initial, chain[-1] if chain else None)
            chain.append(mod)

        # Phase 1 [0, 6]
        for _ in range(0, self._length_phase1):
            _push_node(SmallProfit)

        # Decay 1 [2, 3]
        decay_class = MediumLoss
        for _ in range(0, self._length_decay1):
            _push_node(decay_class)
            decay_class = RapidDecay

        # Phase 2 [1, 6]
        for _ in range(0, self._length_phase2):
            _push_node(SmallProfit)

        # Decay 2 [2, 3]
        decay_class = MediumLoss
        for _ in range(0, self._length_decay2):
            _push_node(decay_class)
            decay_class = RapidDecay

        # Phase 3 [0, 6]
        for _ in range(0, self._length_phase3):
            _push_node(SmallProfit)

        # Build timeline
        assert len(chain) == 12
        for i, mod in enumerate(chain):
            time = TimePeriod(i + 2)
            self.timeline[time] = mod

    @property
    def phases(self) -> List[int]:
        return [
            self._length_phase1,
            self._length_decay1,
            self._length_phase2,
            self._length_decay2,
            self._length_phase3,
        ]

    @property
    def params(self) -> str:
        return str(self.phases)

    @classmethod
    def inner_permutations(cls, initial: int) -> Generator[TripleModel, None, None]:
        for phase1 in range(0, 6 + 1):  # [0, 6] inclusive
            for decay1 in (2, 3):
                for phase2 in range(1, 7 - phase1 + 1):  # [1, 7 - phase1] inclusive
                    yield cls(Price(initial), phase1, decay1, phase2)


class DecayModel(Model):
    _model_class = ModelEnum.decay

    def __init__(self, initial: Price):
        super().__init__(initial)
        mod: Modifier
        parent: Modifier

        # Slice 2: InitialDecay
        mod = InitialDecay(self.initial, None)
        self.timeline[TimePeriod(2)] = mod
        parent = mod

        # Slices 3-13: Decay
        for i in range(3, 14):
            time = TimePeriod(i)
            mod = SlowDecay(self.initial, parent)
            self.timeline[time] = mod
            parent = mod

    @property
    def params(self) -> str:
        # None params with left beef
        return ""


class PeakModel(Model):
    _pattern_latest = 9
    _pattern_earliest: int
    _peak_time: int

    def __init__(self,
                 initial: 'Price',
                 pattern_start: int):

        low = self._pattern_earliest
        high = self._pattern_latest
        if not low <= pattern_start <= high:
            raise ValueError(f"pattern_start must be between {low} and {high}, inclusive")

        super().__init__(initial)
        self._pattern_start = pattern_start
        self._pattern_peak = TimePeriod(pattern_start + self._peak_time)
        self._tail = 9 - pattern_start

    @property
    def peak(self) -> TimePeriod:
        return self._pattern_peak

    @property
    def params(self) -> str:
        return f"peak@{self.peak.name}"

    @classmethod
    def inner_permutations(cls, initial: int) -> Generator[PeakModel, None, None]:
        for patt in range(cls._pattern_earliest, cls._pattern_latest + 1):
            yield cls(Price(initial), patt)


class BumpModel(PeakModel):
    _model_class = ModelEnum.bump
    _pattern_earliest = 2  # Monday AM
    _peak_time = 3         # Fourth price of pattern

    def __init__(self,
                 initial: 'Price',
                 pattern_start: int):

        super().__init__(initial, pattern_start)

        chain: List[Modifier] = []
        decay_class: Type[Modifier]

        def _push_node(mod_cls: Type[Modifier],
                       parent_override: Optional[Modifier] = None) -> Modifier:
            if parent_override is None:
                my_parent = chain[-1] if chain else None
            else:
                my_parent = parent_override
            mod = mod_cls(self.initial, my_parent)
            chain.append(mod)
            return mod

        decay_class = WideLoss
        for _ in range(2, pattern_start):
            _push_node(decay_class)
            decay_class = SlowDecay

        # Pattern starts:
        _push_node(SmallProfit)
        _push_node(SmallProfit)

        # And then gets weird!
        # Create an unlisted parent that represents the peak shared across the next three prices.
        cap = MediumProfit(self.initial, chain[-1])
        _push_node(CappedPassthrough, cap).sub1 = True
        _push_node(Passthrough, cap)
        _push_node(CappedPassthrough, cap).sub1 = True

        # Alright, phew.
        decay_class = WideLoss
        for _ in range(0, self._tail):
            _push_node(decay_class)
            decay_class = SlowDecay

        # Build timeline
        assert len(chain) == 12
        for i, mod in enumerate(chain):
            time = TimePeriod(i + 2)
            self.timeline[time] = mod


class SpikeModel(PeakModel):
    _model_class = ModelEnum.spike
    _pattern_earliest = 3  # Monday PM
    _peak_time = 2         # Third price of pattern

    def __init__(self,
                 initial: 'Price',
                 pattern_start: int):

        super().__init__(initial, pattern_start)

        def _get_pattern(value: int) -> Type[Modifier]:
            # The default:
            cls: Type[Modifier] = SlowDecay

            # Pattern takes priority:
            if value == pattern_start:
                cls = SmallProfit
            elif value == pattern_start + 1:
                cls = MediumProfit
            elif value == pattern_start + 2:
                cls = LargeProfit
            elif value == pattern_start + 3:
                cls = MediumProfit
            elif value == pattern_start + 4:
                cls = SmallProfit
            elif value >= pattern_start + 5:
                # Week finishes out with independent low prices
                cls = WideLoss
            elif value == 2:
                # Normal start-of-week pattern
                cls = InitialDecay

            return cls

        parent = None
        # Slices 2-13: Magic!
        for i in range(2, 14):
            time = TimePeriod(i)
            mod = _get_pattern(i)(self.initial, parent)
            self.timeline[time] = mod
            parent = mod



from __future__ import annotations

import hashlib
import json
import logging
import pickle
from typing import Any, Dict, Generator, Optional

import pydantic
from pydantic import validator
import yaml

from turnips.model import ModelEnum
from turnips.meta import MetaModel
from turnips.multi import MultiModel, BumpModels
from turnips.ttime import TimePeriod

from turnips.plots import plot_models_range_interactive, plot_models_range_data, global_plot


# pylint: disable=too-few-public-methods, no-self-use, no-self-argument

class Base(pydantic.BaseModel):
    class Config:
        extra = 'forbid'


class IslandModel(Base):
    previous_week: ModelEnum = ModelEnum.unknown
    initial_week: bool = False
    timeline: Dict[TimePeriod, Optional[int]]

    @validator('previous_week', pre=True)
    def coerce_model(cls, value: Any) -> Any:
        if isinstance(value, str):
            return ModelEnum[value.lower()]
        return value

    @validator('timeline', pre=True)
    def normalize(cls, value: Any) -> Any:
        if isinstance(value, Dict):
            return {
                TimePeriod.normalize(key): price
                for key, price in value.items()
            }
        return value

    @validator('initial_week')
    def cajole(cls, initial_week: bool, values: Dict[str, Any]) -> Any:
        if values['previous_week'] != ModelEnum.unknown and initial_week:
            raise ValueError("Cannot set initial_week = True when previous_week is set")
        return initial_week

    @property
    def base_price(self) -> Optional[int]:
        return self.timeline.get(TimePeriod.Sunday_AM, None)


class ArchipelagoModel(Base):
    islands: Dict[str, IslandModel]


class Island:
    def __init__(self, name: str, data: IslandModel):
        self._name = name
        self._data = data
        self._models: Optional[MultiModel] = None

    @property
    def name(self) -> str:
        return self._name

    @property
    def previous_week(self) -> ModelEnum:
        return self._data.previous_week

    @property
    def initial_week(self) -> bool:
        return self._data.initial_week

    @property
    def model_group(self) -> MultiModel:
        if self._models is None:
            self.process()
            assert self._models is not None
        return self._models

    def process(self) -> None:
        logging.info(f" == {self.name} island == ")

        base = self._data.base_price
        if self.initial_week:
            self._models = BumpModels()
        else:
            self._models = MetaModel.blank(base)

        logging.info(f"  (%d models)  ", len(self._models))

        for time, price in self._data.timeline.items():
            if price is None:
                continue
            if time.value < TimePeriod.Monday_AM.value:
                continue
            logging.info(f"[{time.name}]: fixing price @ {price}")
            self._models.fix_price(time, price)

    def plot(self):
        return plot_models_range_data(
            self.name,
            list(self.model_group.models),
            self.previous_week
        )

    def checksum(self):
        data = pickle.dumps((self._data.dict(), self.name))
        return hashlib.md5(data).hexdigest()


class Archipelago:
    def __init__(self, data: ArchipelagoModel):
        self._data = data
        self._islands = {name: Island(name, idata) for name, idata
                         in self._data.islands.items()}

    @classmethod
    def parse_obj(cls, obj: Dict[Any, Any]) -> Archipelago:
        return cls(ArchipelagoModel.parse_obj(obj))

    @classmethod
    def load_json(cls, data: str) -> Archipelago:
        jdata = json.loads(data)
        return cls.parse_obj(jdata)

    @classmethod
    def load_yaml(cls, data: str) -> Archipelago:
        ydata = yaml.safe_load(data)
        return cls.parse_obj(ydata)

    @classmethod
    def load_file(cls, filename: str) -> Archipelago:
        if '.yml' in filename or '.yaml' in filename:
            with open(filename, "r") as infile:
                ydoc = yaml.safe_load(infile)
                return cls.parse_obj(ydoc)

        return cls(ArchipelagoModel.parse_file(filename))

    @property
    def groups(self) -> Generator[MultiModel, None, None]:
        for island in self._islands.values():
            yield island.model_group

    @property
    def islands(self) -> Generator[Island, None, None]:
        for island in self._islands.values():
            yield island

    def summary(self) -> None:
        for island in self.islands:
            print(f"{island.name}")
            print('-' * len(island.model_group))
            print('')
            island.model_group.report()
            print('')

        # The initial doesn't matter here, it's ignored.
        print('Archipelago Summary')
        print('-' * 80)
        archipelago = MetaModel(-1, self.groups)
        archipelago.summary()
        print('-' * 80)

    def plot(self) -> None:
        for island in self.islands:
            plot_models_range_interactive(island.name,
                                          list(island.model_group.models),
                                          island.previous_week)
        global_plot(self.groups)



#!/usr/bin/env python3

from __future__ import annotations

import enum
from typing import Union


AnyTime = Union[str, int, 'TimePeriod']

class TimePeriod(enum.Enum):
    Sunday_AM = 0
    Sunday_PM = 1
    Monday_AM = 2
    Monday_PM = 3
    Tuesday_AM = 4
    Tuesday_PM = 5
    Wednesday_AM = 6
    Wednesday_PM = 7
    Thursday_AM = 8
    Thursday_PM = 9
    Friday_AM = 10
    Friday_PM = 11
    Saturday_AM = 12
    Saturday_PM = 13

    @classmethod
    def normalize(cls, value: AnyTime) -> TimePeriod:
        if isinstance(value, TimePeriod):
            return value
        if isinstance(value, int):
            return cls(value)
        return cls[value]



#!/usr/bin/env python3

from __future__ import annotations

import math
from typing import Callable, Optional, Tuple


class Price:
    def __init__(self,
                 price: float,
                 upper: Optional[float] = None,
                 ufilter: Optional[Callable[[float], float]] = None):
        self._lower = price
        self._upper = price if upper is None else upper
        self._user_filter = ufilter

    def _filter(self, value: float) -> float:
        if self._user_filter is not None:
            return self._user_filter(math.ceil(value))
        return math.ceil(value)

    @property
    def is_atomic(self) -> bool:
        return self._lower == self._upper

    @property
    def lower(self) -> float:
        return self._filter(self._lower)

    @property
    def upper(self) -> float:
        return self._filter(self._upper)

    @property
    def value(self) -> float:
        if not self.is_atomic:
            raise ValueError("No single value available; this is a range")
        return self.lower

    @property
    def raw(self) -> Tuple[float, float]:
        return (self._lower, self._upper)

    def __repr__(self) -> str:
        if self.is_atomic:
            return '{}({})'.format(self.__class__.__name__, self._lower)
        return '{}({}, {})'.format(self.__class__.__name__, self._lower, self._upper)

    def __str__(self) -> str:
        if self.is_atomic:
            return str(self.value)
        return f"[{self.lower}, {self.upper}]"



#!/usr/bin/env python3

from __future__ import annotations

from decimal import Decimal, ROUND_DOWN, ROUND_UP
import logging
import math
from typing import Optional

from turnips.price import Price

def find_lower_bound(value: int, base: float, precision: str = '0.0001') -> float:
    """
    This awful function computes the smallest float x such that:
    ceil(x * base) == value
    to within some arbitrary quantized precision, e.g. 0.0001.
    """
    lower_int = value - 1
    lower_bound = Decimal(lower_int / base)
    lower_approx = float(lower_bound.quantize((Decimal(precision)),
                                              rounding=ROUND_DOWN))
    while math.ceil(lower_approx * base) < value:
        lower_approx += 0.0001

    return lower_approx


def find_upper_bound(value: int, base: float, precision: str = '0.0001') -> float:
    """
    This awful function computes the largest float x such that:
    ceil(x * base) == value
    to within some arbitrary quantized precision, e.g. 0.0001.
    """
    upper_bound = Decimal(value / base)
    upper_approx = float(upper_bound.quantize((Decimal(precision)),
                                              rounding=ROUND_UP))
    while math.ceil(upper_approx * base) > value:
        upper_approx -= 0.0001

    return upper_approx


class Modifier:
    def __init__(self,
                 base: Price,
                 parent: Optional[Modifier] = None):
        self._base = base
        self._parent = parent
        self._exact_price: Optional[float] = None
        self._static_low: Optional[float] = None
        self._static_high: Optional[float] = None
        self.sub1 = False

    def __str__(self) -> str:
        low = self.lower_bound
        high = self.upper_bound
        return f"x[{low:0.4f}, {high:0.4f}]"

    @property
    def name(self) -> str:
        return self.__class__.__name__

    @property
    def parent(self) -> Modifier:
        if self._parent is None:
            raise ValueError("Modifier node has no parent")
        return self._parent

    def _default_lower_bound(self) -> float:
        raise NotImplementedError

    @property
    def lower_bound(self) -> float:
        if self._static_low is not None:
            return self._static_low
        return self._default_lower_bound()

    def _default_upper_bound(self) -> float:
        raise NotImplementedError

    @property
    def upper_bound(self) -> float:
        if self._static_high is not None:
            return self._static_high
        return self._default_upper_bound()

    @property
    def price(self) -> Price:
        if self._exact_price is not None:
            return Price(self._exact_price)

        lower = self.lower_bound * self._base.lower
        upper = self.upper_bound * self._base.upper

        if self.sub1:
            return Price(lower, upper, ufilter=lambda f: f - 1)

        return Price(lower, upper)

    def tighten_bounds(self,
                       lower: Optional[float] = None,
                       upper: Optional[float] = None) -> None:
        if lower is not None and upper is not None:
            if lower > upper:
                raise ValueError("Invalid Argument (Range is reversed): [{lower}, {upper}]")

        if lower is not None:
            if self.lower_bound <= lower <= self.upper_bound:
                self._static_low = lower

        if upper is not None:
            if self.lower_bound <= upper <= self.upper_bound:
                self._static_high = upper

    def fix_price(self, price: int) -> None:
        # Protection against typos and fat fingers:
        if self._exact_price is not None:
            if self._exact_price != price:
                raise ValueError("Invalid Argument: Cannot re-fix price range with new price")

        price_window = self.price

        # Check that the new price isn't immediately outside of what we presently consider possible
        if price < price_window.lower:
            msg = f"Cannot fix price at {price:d}, below model minimum {price_window.lower:d}"
            raise ArithmeticError(msg)
        if price > price_window.upper:
            msg = f"Cannot fix price at {price:d}, above model maximum {price_window.upper:d}"
            raise ArithmeticError(msg)

        # Calculate our presently understood modifier bounds
        current_mod_low = self.lower_bound
        current_mod_high = self.upper_bound
        logging.debug('current modifier range: [{:0.4f}, {:0.4f}]'.format(
            current_mod_low, current_mod_high))

        # Sometimes a price given is (modifier * price) - 1;
        # to compute the correct coefficients, we need to add one!
        if self.sub1:
            unfiltered_price = price + 1
        else:
            unfiltered_price = price

        # For the price given, determine the modifiers that could have produced that value.
        #
        # NB: Accommodate an unknown base price by computing against
        # the lower and upper bounds of the base price.
        #
        # NB2: The boundaries here can change over time, depending on a few things:
        #
        # (1) If the base price range is refined in the future, this
        #     calculation could change.  (At present, MultiModel
        #     always builds models with fixed base prices, not
        #     ranges.)
        #
        # (2) If we are a dynamic modifier node and any of our parents
        #     refine THEIR bounds, this could change.  This is
        #     generally only a problem when there are gaps in the
        #     data; Subsequent constraints will be necessarily less
        #     accurate.
        #
        # FIXME: Once this computation is performed, it's completely
        #        static. Oops!  This code needs to be a little more
        #        dynamically adaptable.

        modifiers = []
        for bound in (self._base.lower, self._base.upper):
            modifiers.append(find_lower_bound(unfiltered_price, bound))
            modifiers.append(find_upper_bound(unfiltered_price, bound))
        new_mod_low = min(modifiers)
        new_mod_high = max(modifiers)
        logging.debug('fixed modifier range: [{:0.4f}, {:0.4f}]'.format(new_mod_low, new_mod_high))

        # If our modifiers are out of scope, we can reject the price for this model.
        if new_mod_low > current_mod_high:
            msg = 'New low modifier ({:0.4f}) out of range [{:0.4f}, {:0.4f}]'.format(
                new_mod_low, current_mod_low, current_mod_high)
            raise ArithmeticError(msg)
        if new_mod_high < current_mod_low:
            msg = 'New high modifier ({:0.4f}) out of range [{:0.4f}, {:0.4f}]'.format(
                new_mod_high, current_mod_low, current_mod_high)
            raise ArithmeticError(msg)
        assert new_mod_low <= new_mod_high, 'God has left us'

        # Our calculated modifier range might overlap with the base model range.
        # Base: [    LxxxxxxH      ]
        # Calc: [ lxxxxxh          ]
        # Calc: [        lxxxxh    ]
        #
        # i.e. our low end may be lower then the existing low end,
        # or the high end might be higher then the existing high end.
        #
        # The assertions in the code block above only assert that:
        # (1) l <= H
        # (2) h >= L
        #
        # Which means that these formulations are valid:
        # Base: [    LxxxxxxH      ]
        # Calc: [           lxxxxh ]
        # Calc: [ lxxh             ]
        #
        # This is fine, it just means that we can rule out some of the modifiers
        # in the range of possibilities.
        # Clamp the new computed ranges to respect the original boundaries.
        new_mod_low = max(new_mod_low, current_mod_low)
        new_mod_high = min(new_mod_high, current_mod_high)
        logging.debug('clamped modifier range: [{:0.4f}, {:0.4f}]'.format(
            new_mod_low, new_mod_high))

        self.tighten_bounds(new_mod_low, new_mod_high)
        self._exact_price = price


class RootModifier(Modifier):
    def _default_lower_bound(self) -> float:
        return 0.00

    def _default_upper_bound(self) -> float:
        return 6.00


class StaticModifier(Modifier):
    default_lower = 0.00
    default_upper = 1.00

    def _default_lower_bound(self) -> float:
        return self.default_lower

    def _default_upper_bound(self) -> float:
        return self.default_upper


class InitialDecay(StaticModifier):
    default_lower = 0.85
    default_upper = 0.90


class WideLoss(StaticModifier):
    default_lower = 0.40
    default_upper = 0.90


class MediumLoss(StaticModifier):
    default_lower = 0.6
    default_upper = 0.8


class SmallProfit(StaticModifier):
    default_lower = 0.90
    default_upper = 1.40


class MediumProfit(StaticModifier):
    default_lower = 1.4
    default_upper = 2.0


class LargeProfit(StaticModifier):
    default_lower = 2.0
    default_upper = 6.0


class DynamicModifier(Modifier):
    """
    DynamicModifier inherits the modifier coefficients of its parent.
    It can modify these coefficients with two deltas.
    """
    delta_lower = 0.00
    delta_upper = 0.00

    def _default_lower_bound(self) -> float:
        return self.parent.lower_bound + self.delta_lower

    def _default_upper_bound(self) -> float:
        return self.parent.upper_bound + self.delta_upper


class Passthrough(DynamicModifier):
    @property
    def name(self) -> str:
        return f"{self.parent.name}*"

    def fix_price(self, price: int) -> None:
        self.parent.fix_price(price)
        super().fix_price(price)


class CappedPassthrough(DynamicModifier):
    # Inherit the *static* lower bound of our parent
    def _default_lower_bound(self) -> float:
        assert isinstance(self.parent, StaticModifier)
        return self.parent.default_lower

    # Upper bound is our parent's dynamic upper bound, as normal for Dynamic nodes

    def fix_price(self, price: int) -> None:
        super().fix_price(price)
        self.parent.tighten_bounds(lower=self.lower_bound)

    @property
    def name(self) -> str:
        return f"{self.parent.name} (Capped)"


class SlowDecay(DynamicModifier):
    delta_lower = -0.05
    delta_upper = -0.03


class RapidDecay(DynamicModifier):
    delta_lower = -0.10
    delta_upper = -0.04



#!/usr/bin/env python3

from __future__ import annotations

from typing import Generator, Iterable, Optional

from turnips.model import Model
from turnips.multi import (
    MultiModel,
    TripleModels,
    SpikeModels,
    DecayModels,
    BumpModels,
)
from turnips.ttime import AnyTime


class MetaModel(MultiModel):
    # This is either a horrid abuse of OO, or brilliant.
    # We override only methods that refer to self._models directly;
    # otherwise, self.models takes care of it.

    def __init__(self,
                 initial: Optional[int],
                 groups: Iterable[MultiModel]):
        super().__init__(initial)
        self._children = list(groups)

    @classmethod
    def blank(cls, initial: Optional[int] = None) -> MetaModel:
        groups = [
            TripleModels(initial),
            SpikeModels(initial),
            DecayModels(initial),
            BumpModels(initial),
        ]
        return cls(initial, groups)

    @property
    def models(self) -> Generator[Model, None, None]:
        for child in self._children:
            yield from child.models

    def fix_price(self, time: AnyTime, price: int) -> None:
        for child in self._children:
            child.fix_price(time, price)

    def __bool__(self) -> bool:
        return len(self) != 0

    def __len__(self) -> int:
        return sum(len(child) for child in self._children)

    def detailed_report(self) -> None:
        self._children = list(filter(None, self._children))

        if not self._children:
            print("--- No Viable Model Groups ---")
            return

        if len(self._children) == 1:
            self._children[0].report(show_summary=False)
            return

        print("Meta-Analysis: ")
        print("total possible models: {}".format(len(self)))
        print('----------')
        for child in self._children:
            if child:
                child.report()
                print('----------')



#!/usr/bin/env python3

from turnips import archipelago

EXAMPLE_DATA = '''
{
    "islands": {
        "trimex": {
            "previous_week": "bump",
            "timeline": {
                "Sunday_AM":   110,
                "Monday_AM":   58,
                "Monday_PM":   null,
                "Tuesday_AM":  null,
                "Tuesday_PM":  null,
                "Wednesday_AM":null,
                "Wednesday_PM":null,
                "Thursday_AM": null,
                "Thursday_PM": null,
                "Friday_AM":   null,
                "Friday_PM":   null,
                "Saturday_AM": null,
                "Saturday_PM": null
            }
        },
        "Aduien": {
            "previous_week": "bump",
            "timeline": {
                "Sunday_AM":   91,
                "Monday_AM":   113,
                "Monday_PM":   106,
                "Tuesday_AM":  98,
                "Tuesday_PM":  null,
                "Wednesday_AM":null,
                "Wednesday_PM":null,
                "Thursday_AM": null,
                "Thursday_PM": null,
                "Friday_AM":   null,
                "Friday_PM":   null,
                "Saturday_AM": null,
                "Saturday_PM": null
            }
        },
        "Torishima": {
            "previous_week": "spike",
            "timeline": {
                "Sunday_AM":   106,
                "Monday_AM":   105,
                "Monday_PM":   134,
                "Tuesday_AM":  121,
                "Tuesday_PM":  null,
                "Wednesday_AM":null,
                "Wednesday_PM":null,
                "Thursday_AM": null,
                "Thursday_PM": null,
                "Friday_AM":   null,
                "Friday_PM":   null,
                "Saturday_AM": null,
                "Saturday_PM": null
            }
        },
        "File Isle": {
            "previous_week": "bump",
            "timeline": {
                "Sunday_AM":   96,
                "Monday_AM":   83,
                "Monday_PM":   79,
                "Tuesday_AM":  null,
                "Tuesday_PM":  null,
                "Wednesday_AM":null,
                "Wednesday_PM":null,
                "Thursday_AM": null,
                "Thursday_PM": null,
                "Friday_AM":   null,
                "Friday_PM":   null,
                "Saturday_AM": null,
                "Saturday_PM": null
            }
        },
        "Kibo": {
            "previous_week": "bump",
            "timeline": {
                "Sunday_AM":   93,
                "Monday_AM":   73,
                "Monday_PM":   69,
                "Tuesday_AM":  65,
                "Tuesday_PM":  null,
                "Wednesday_AM":null,
                "Wednesday_PM":null,
                "Thursday_AM": null,
                "Thursday_PM": null,
                "Friday_AM":   null,
                "Friday_PM":   null,
                "Saturday_AM": null,
                "Saturday_PM": null
            }
        },
        "Calidris": {
            "previous_week": "bump",
            "timeline": {
                "Sunday_AM":   99,
                "Monday_AM":   98,
                "Monday_PM":   93,
                "Tuesday_AM":  null,
                "Tuesday_PM":  null,
                "Wednesday_AM":null,
                "Wednesday_PM":null,
                "Thursday_AM": null,
                "Thursday_PM": null,
                "Friday_AM":   null,
                "Friday_PM":   null,
                "Saturday_AM": null,
                "Saturday_PM": null
            }
        },
        "Harvest": {
            "previous_week": "bump",
            "timeline": {
                "Sunday_AM":   110,
                "Monday_AM":   81,
                "Monday_PM":   77,
                "Tuesday_AM":  73,
                "Tuesday_PM":  68,
                "Wednesday_AM":null,
                "Wednesday_PM":null,
                "Thursday_AM": null,
                "Thursday_PM": null,
                "Friday_AM":   null,
                "Friday_PM":   null,
                "Saturday_AM": null,
                "Saturday_PM": null
            }
        },
        "Wolfshire": {
            "initial_week": true,
            "timeline": {
                "Sunday_AM":   107,
                "Monday_AM":   95,
                "Monday_PM":   92,
                "Tuesday_AM":  88,
                "Tuesday_PM":  null,
                "Wednesday_AM":null,
                "Wednesday_PM":null,
                "Thursday_AM": null,
                "Thursday_PM": null,
                "Friday_AM":   null,
                "Friday_PM":   null,
                "Saturday_AM": null,
                "Saturday_PM": null
            }
        },
        "covid19": {
            "initial_week": true,
            "timeline": {
                "Sunday_AM":   97,
                "Monday_AM":   null,
                "Monday_PM":   null,
                "Tuesday_AM":  70,
                "Tuesday_PM":  null,
                "Wednesday_AM":null,
                "Wednesday_PM":null,
                "Thursday_AM": null,
                "Thursday_PM": null,
                "Friday_AM":   null,
                "Friday_PM":   null,
                "Saturday_AM": null,
                "Saturday_PM": null
            }
        }
    }
}
'''

def main() -> None:
    # See also `turnips.py --plot sample.json`
    islands = archipelago.Archipelago.load_json(EXAMPLE_DATA)
    islands.plot()


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

from __future__ import annotations

import argparse
import logging

from turnips import archipelago

def main() -> None:
    parser = argparse.ArgumentParser(description='Gaze into the runes')
    parser.add_argument('archipelago', metavar='archipelago-json',
                        help='JSON file with turnip price data for the week')
    parser.add_argument('--plot', dest='plot', action='store_true',
                        help='Draw some pretty graphs')
    parser.add_argument('--verbose', dest='verbose', action='store_true',
                        help='Turn on some informational messages')
    parser.add_argument('--debug', dest='debug', action='store_true',
                        help='Turn on debugging messages')
    args = parser.parse_args()

    if args.debug:
        logging.basicConfig(level=logging.DEBUG)
    elif args.verbose:
        logging.basicConfig(level=logging.INFO)

    arch = archipelago.Archipelago.load_file(args.archipelago)
    arch.summary()
    if args.plot:
        arch.plot()


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

from __future__ import annotations

from collections import Counter
import logging
from typing import Callable, Dict, Generator, List, Optional, Set, Type
from typing import Counter as CounterType

from turnips.model import (
    Model,
    TripleModel,
    SpikeModel,
    DecayModel,
    BumpModel,
)
from turnips.ttime import AnyTime, TimePeriod


class RangeSet:
    def __init__(self) -> None:
        self._set: Set[int] = set()

    def add(self, value: int) -> None:
        self._set.add(value)

    def __str__(self) -> str:
        sortlist = sorted(list(self._set))
        ranges = []

        seed = sortlist.pop(0)
        current = [seed, seed]

        for item in sortlist:
            if item == current[1] + 1:
                current[1] = item
            else:
                ranges.append((current[0], current[1]))
                current = [item, item]
        ranges.append((current[0], current[1]))

        chunks = []
        for prange in ranges:
            if prange[0] == prange[1]:
                chunk = f"{prange[0]}"
            else:
                chunk = f"[{prange[0]}, {prange[1]}]"
            chunks.append(chunk)

        if len(chunks) == 1:
            return str(chunks[0])
        return "{" + ", ".join(chunks) + "}"


class MultiModel:
    """
    A collection of several models that we can aggregate data about.
    """
    def __init__(self, initial: Optional[int] = None):
        self._initial = initial
        self._models: Dict[int, Model] = {}

    @property
    def models(self) -> Generator[Model, None, None]:
        for model in self._models.values():
            yield model

    def fix_price(self, time: AnyTime, price: int) -> None:
        if not self:
            assert RuntimeError("No viable models to fix prices on!")

        timeslice = TimePeriod.normalize(time)
        remove_queue = []
        for index, model in self._models.items():
            try:
                model.fix_price(timeslice, price)
            except ArithmeticError as exc:
                logging.info(f"Ruled out model: {model.name}")
                logging.debug(f"  Reason: {timeslice.name} price={price} not possible:")
                logging.debug(f"    {str(exc)}")
                remove_queue.append(index)

        for i in remove_queue:
            del self._models[i]

    def __bool__(self) -> bool:
        return bool(self._models)

    def __len__(self) -> int:
        return len(self._models)

    def __str__(self) -> str:
        if not self:
            return "--- No Viable Models ---"
        return "\n\n".join([str(model) for model in self.models])

    def histogram(self) -> Dict[str, CounterType[int]]:
        histogram: Dict[str, CounterType[int]] = {}
        for model in self.models:
            mhist = model.histogram()
            for timename in mhist:
                counter = histogram.setdefault(timename, Counter())
                counter.update(mhist[timename])
        return histogram

    def summary(self) -> None:
        print('')
        print("  Summary: ")
        print("    {:13} {:23} {:23} {:6}".format('Time', 'Price', 'Likely', 'Odds'))
        hist = self.histogram()
        for time, pricecounts in hist.items():
            # Gather possible prices
            pset = RangeSet()
            for price in pricecounts.keys():
                pset.add(price)

            # Determine likeliest price(s)
            n_possibilities = sum(pricecounts.values())
            likeliest = max(pricecounts.items(), key=lambda x: x[1])
            likelies = list(filter(lambda x: x[1] >= likeliest[1], pricecounts.items()))

            sample_size = len(likelies) * likeliest[1]
            pct = 100 * (sample_size / n_possibilities)

            rset = RangeSet()
            for likely in likelies:
                rset.add(likely[0])

            time_col = f"{time}:"
            price_col = f"{str(pset)};"
            likely_col = f"{str(rset)};"
            chance_col = f"({pct:0.2f}%)"
            print(f"    {time_col:13} {price_col:23} {likely_col:23} {chance_col:6}")

    def detailed_report(self) -> None:
        raise NotImplementedError

    def report(self, show_summary: bool = True) -> None:
        if len(self) == 1:
            for model in self.models:
                print(model)
        else:
            self.detailed_report()
            if show_summary:
                self.summary()

    def chatty_fix_price(self, time: AnyTime, price: int) -> None:
        timeslice = TimePeriod.normalize(time)
        self.fix_price(time, price)
        print(f"Added {timeslice.name} @ {price};")
        self.report()


class PeakModels(MultiModel):
    """
    PeakModels gathers aggregate data about Bump and Spike models.
    """
    _interior_class: Type[Model]

    def __init__(self, initial: Optional[int] = None):
        super().__init__(initial)
        for i, model in enumerate(self._interior_class.permutations(initial)):
            self._models[i] = model

    def detailed_report(self) -> None:
        buckets: Dict[TimePeriod, List[Model]] = {}
        prices: Dict[TimePeriod, RangeSet] = {}
        global_prices = RangeSet()
        for model in self._models.values():
            assert isinstance(model, (BumpModel, SpikeModel))  # Shut up mypy, I know it's bad OO
            peak = model.peak
            buckets.setdefault(peak, []).append(model)
            prices.setdefault(peak, RangeSet()).add(int(model.initial.value))
            global_prices.add(int(model.initial.value))

        print(f"{self._interior_class.__name__} Model Analyses:")
        print(f"  {len(self._models)} model(s) remaining")
        print(f"  base price: {str(global_prices)}")

        if len(buckets) != 1:
            print('')
            print(f"  {len(buckets)} possible peak times:")

        for key, _models in buckets.items():
            groupprices = prices[key]
            bullet = '' if len(buckets) == 1 else '- '
            print(f"  {bullet}peak time: {key}")
            if len(buckets) == 1:
                # Don't re-print the prices for just one group.
                continue
            print(f"      base price: {str(groupprices)}")


class BumpModels(PeakModels):
    _interior_class = BumpModel


class SpikeModels(PeakModels):
    _interior_class = SpikeModel


class DecayModels(MultiModel):
    def __init__(self, initial: Optional[int] = None):
        super().__init__(initial)
        for i, model in enumerate(DecayModel.permutations(initial)):
            self._models[i] = model

    def detailed_report(self) -> None:
        global_prices = RangeSet()
        for model in self._models.values():
            global_prices.add(int(model.initial.value))

        print(f"Decay Model Analyses:")
        print(f"  {len(self._models)} model(s) remaining")
        print(f"  base price: {str(global_prices)}")


class TripleModels(MultiModel):
    def __init__(self, initial: Optional[int] = None):
        super().__init__(initial)
        for i, model in enumerate(TripleModel.permutations(initial)):
            self._models[i] = model

    def detailed_report(self) -> None:
        # Alright, long story short:
        # This method (and _analyze_models) loop over the known parameters (params)
        # and their values, and look for parameters which are now "fixed" in the remaining data.
        #
        # If any are found to have only one possible value, we print that value and
        # delete it from the dict.
        #
        # If there are parameters with multiple values, loop over each value
        # and recursively call _analyze_models on just those models.

        params = {
            'phase1': lambda m: m.phases[0],
            'decay1': lambda m: m.phases[1],
            'phase2': lambda m: m.phases[2],
            'price': lambda m: int(m.initial.value),
        }

        print(f"Triple Model Analyses:")
        print(f"  {len(self._models)} model(s) remaining")

        mlist: List[TripleModel] = []
        for model in self._models.values():
            assert isinstance(model, TripleModel)
            mlist.append(model)

        self._analyze_models(mlist, params)

    def _analyze_models(self,
                        models: List[TripleModel],
                        params: Dict[str, Callable[[TripleModel], int]],
                        indent: int = 2) -> None:
        indent_str = ' ' * indent

        buckets: Dict[str, Dict[int, List[TripleModel]]] = {}
        for model in models:
            for param, param_fn in params.items():
                pdict = buckets.setdefault(param, {})
                pvalue = param_fn(model)
                pdict.setdefault(pvalue, []).append(model)

        remove_queue = []
        for param, pvalues in buckets.items():
            if len(pvalues) == 1:
                pvalue = list(pvalues.keys())[0]
                print(f"{indent_str}{param}: {pvalue}")
                remove_queue.append(param)

        for remove in remove_queue:
            buckets.pop(remove)
            params.pop(remove)

        if len(buckets) == 1:
            # Only one parameter left, so just print it now, as a list.
            pset = RangeSet()
            for param, pvalues in buckets.items():
                for pvalue in pvalues.keys():
                    pset.add(pvalue)
            print(f"{indent_str}{param}: {str(pset)}")
            return

        for param, pvalues in buckets.items():
            params.pop(param)
            for pvalue in pvalues.keys():
                print(f"{indent_str}- {param}: {pvalue}")
                pcopy = params.copy()
                self._analyze_models(pvalues[pvalue], pcopy, indent + 2)
            return



import dunamai as _dunamai


def version_from_git():
    try:
        choice = _dunamai.Version.from_any_vcs
        version = _dunamai.get_version("turbot", first_choice=choice).serialize()
        return version if version is not None and version != "0.0.0" else None
    except RuntimeError:
        return None


def version_from_package():
    try:
        choice = _dunamai.Version.from_any_vcs
        version = _dunamai.get_version("turbot", third_choice=choice).serialize()
        return version if version is not None and version != "0.0.0" else None
    except RuntimeError:
        return None


def version_from_toml():
    import toml
    from os.path import dirname, realpath
    from pathlib import Path

    try:
        PKG_ROOT = dirname(realpath(__file__))
        SRC_ROOT = Path(PKG_ROOT).parent
        REPO_ROOT = SRC_ROOT.parent
        pyproject = toml.load(REPO_ROOT / "pyproject.toml")
    except:
        return None

    tool = pyproject.get("tool", {})
    poetry = tool.get("poetry", {})
    version = poetry.get("version", None)
    return version if version is not None and version != "0.0.0" else None


# Below we attempt to detect the version from git first, and if
# that fails we fallback to getting it from the project, and if
# that fails we'll just try and read it directly from pyproject.toml.
# If we can successfully get it from git, we can get a development
# version name like 5.0.2.post2.dev0+1d15510 for example.

__version__ = version_from_git()
if __version__ is None:
    __version__ = version_from_package()
if __version__ is None:
    __version__ = version_from_toml()
if __version__ is None:
    __version__ = "unknown"



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['turbot', 'turnips']

package_data = \
{'': ['*'], 'turbot': ['data/*']}

install_requires = \
['click>=7.1.2,<8.0.0',
 'discord-py>=1.3.3,<2.0.0',
 'dunamai>=1.1.0,<2.0.0',
 'humanize>=2.4.0,<3.0.0',
 'hupper>=1.10.2,<2.0.0',
 'matplotlib>=3.2.1,<4.0.0',
 'numpy>=1.18.4,<2.0.0',
 'pandas>=1.0.3,<2.0.0',
 'pydantic>=1.5.1,<2.0.0',
 'python-dateutil>=2.8.1,<3.0.0',
 'pytz>=2020.1,<2021.0',
 'pyyaml>=5.3.1,<6.0.0',
 'unidecode>=1.1.1,<2.0.0']

entry_points = \
{'console_scripts': ['turbot = turbot:main']}

setup_kwargs = {
    'name': 'turbot',
    'version': '6.2.0',
    'description': 'Provides a Discord client and utilities for everything Animal Crossing: New Horizons.',
    'long_description': '<img align="right" src="https://raw.githubusercontent.com/theastropath/turbot/master/turbot.png" />\n\n# Turbot\n\n[![build][build-badge]][build]\n[![pypi][pypi-badge]][pypi]\n[![python][python-badge]][python]\n[![codecov][codecov-badge]][codecov]\n[![black][black-badge]][black]\n[![mit][mit-badge]][mit]\n\nA Discord bot for everything _Animal Crossing: New Horizons_.\n\n![screenshot](https://user-images.githubusercontent.com/1903876/80841531-787c2f00-8bb4-11ea-8975-cc619b978635.png)\n\n## 🤖 Running the bot\n\nFirst install `turbot` using [`pip`](https://pip.pypa.io/en/stable/):\n\n```shell\npip install turbot\n```\n\nThen you must configure two things:\n\n1. Your Discord bot token.\n2. The list of channels you want `turbot` to monitor.\n\nTo provide your Discord bot token either set an environment variable named\n`TURBOT_TOKEN` to the token or paste it into a file named `token.txt`.\n\nFor the list of channels you can provide channel names on the command line using\nany number of `--channel "name"` options. Alternatively you can create a file\nnamed `channels.txt` where each line of the file is a channel name. You can\nalso specify them via the environment variable `TURBOT_CHANNELS` as a semicolon\ndelimited string, for example: `export TURBOT_CHANNELS="some;channels;here"`.\n\nMore usage help can be found by running `turbot --help`.\n\n## 📱 Using the bot\n\nOnce you\'ve connected the bot to your server, you can interact with it over\nDiscord via the following commands in any of the authorized channels.\n\n- `!about`: Get information about Turbot\n- `!help`: Provides detailed help about all of the following commands\n\n### 🤔 User Preferences\n\nThese commands allow users to set their preferences. These preferences are used\nto make other commands more relevant, for example by converting times to the\nuser\'s preferred timezone.\n\n- `!info`: Get a user\'s information\n- `!pref`: Set a user preference; use command to get a list of available options\n\n### 💸 Turnips\n\nThese commands help users buy low and sell high in the stalk market.\n\n- `!best`: Look for the current best sell or buy\n- `!buy`: Save a buy price\n- `!clear`: Clear your price data\n- `!graph`: Graph price data\n- `!history`: Get price history\n- `!lastweek`: Get graph for last week\'s price data\n- `!oops`: Undo the last price data\n- `!predict`: Predict your price data for the rest of the week\n- `!reset`: Reset all users\' data\n- `!sell`: Save a sell price\n\n### 🐟 Fish & 🐞 Bugs\n\nProvides users with information on where and when to catch critters.\n\n- `!bugs`: Get information on bugs\n- `!fish`: Get information on fish\n- `!new`: Get information on newly available fish and bugs\n\n### 🦴 Fossils, 🖼️ Art, & 🎶 Songs\n\nWhen a community of users tracks collectables and trades them between each\nother, everyone finishes collecting everything in the game so much more quickly\nthan they would on their own.\n\nThese commands can also help users tell fake art from real art.\n\n- `!art`: Get information on an art piece\n- `!collect`: Make something as collected\n- `!collected`: Show the things you\'ve collected so far\n- `!count`: Count the number of collected things you have\n- `!needed`: Find out what collectables are needed by you and others\n- `!search`: Search for someone who needs a fossil or art\n- `!uncollect`: Remove something from your collection\n- `!uncollected`: Get a list of things that you haven\'t collected yet\n\n### ⚛️ Heroku Support\n\nTurbot supports deployment to Heroku out of the box.\nSee our documentation on [Heroku Support](HEROKU.md) for details.\n\n---\n\n[MIT][mit] © [TheAstropath][theastropath], [lexicalunit][lexicalunit] et [al][contributors]\n\n[black-badge]:      https://img.shields.io/badge/code%20style-black-000000.svg\n[black]:            https://github.com/psf/black\n[build-badge]:      https://github.com/theastropath/turbot/workflows/build/badge.svg\n[build]:            https://github.com/theastropath/turbot/actions\n[codecov-badge]:    https://codecov.io/gh/theastropath/turbot/branch/master/graph/badge.svg\n[codecov]:          https://codecov.io/gh/theastropath/turbot\n[contributors]:     https://github.com/theastropath/turbot/graphs/contributors\n[lexicalunit]:      http://github.com/lexicalunit\n[mit-badge]:        https://img.shields.io/badge/License-MIT-yellow.svg\n[mit]:              https://opensource.org/licenses/MIT\n[python-badge]:     https://img.shields.io/badge/python-3.7+-blue.svg\n[python]:           https://www.python.org/\n[theastropath]:     https://github.com/theastropath\n[pypi]:             https://pypi.org/project/turbot/\n[pypi-badge]:       https://img.shields.io/pypi/v/turbot\n',
    'author': 'TheAstropath',
    'author_email': 'theastropath@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/theastropath/turbot',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4',
}


setup(**setup_kwargs)



#!/usr/bin/env python3

from __future__ import annotations

from typing import Dict, Union

from turnips.model import ModelEnum, TRIPLE, SPIKE, DECAY, BUMP, UNKNOWN

MARKOV: Dict[Union[None, ModelEnum], Dict[ModelEnum, float]] = {
    TRIPLE: {
        TRIPLE:  0.2,
        SPIKE:   0.3,
        DECAY:   0.15,
        BUMP:    0.35,
    },
    SPIKE: {
        TRIPLE:  0.5,
        SPIKE:   0.05,
        DECAY:   0.2,
        BUMP:    0.25,
    },
    DECAY: {
        TRIPLE:  0.25,
        SPIKE:   0.45,
        DECAY:   0.05,
        BUMP:    0.25,
    },
    BUMP: {
        TRIPLE:  0.45,
        SPIKE:   0.25,
        DECAY:   0.15,
        BUMP:    0.15,
    },
    UNKNOWN: {
        TRIPLE:  0.25,
        SPIKE:   0.25,
        DECAY:   0.25,
        BUMP:    0.25,
    },
    None: {
        BUMP:    1.0,
    },
}



import inspect
import json
import logging
import random
import re
import sys
from collections import defaultdict
from contextlib import redirect_stdout
from datetime import datetime, timedelta
from itertools import product
from os import getenv
from pathlib import Path
from subprocess import run

import click
import discord
import hupper
import matplotlib
import matplotlib.dates as mdates
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import pytz
from dateutil.relativedelta import relativedelta
from humanize import naturaltime
from unidecode import unidecode

from turbot._version import __version__
from turbot.assets import Assets, s
from turbot.data import Data
from turnips.archipelago import Archipelago
from turnips.plots import plot_models_range

matplotlib.use("Agg")

RUNTIME_ROOT = Path(".")
SCRIPTS_DIR = RUNTIME_ROOT / "scripts"
DEFAULT_DB_DIR = RUNTIME_ROOT / "db"
DEFAULT_CONFIG_TOKEN = RUNTIME_ROOT / "token.txt"
DEFAULT_CONFIG_CHANNELS = RUNTIME_ROOT / "channels.txt"
TMP_DIR = RUNTIME_ROOT / "tmp"
GRAPHCMD_FILE = TMP_DIR / "graphcmd.png"
LASTWEEKCMD_FILE = TMP_DIR / "lastweek.png"
MIGRATIONS_FILE = DEFAULT_DB_DIR / "migrations.txt"
MIGRATIONS_DIR = SCRIPTS_DIR / "migrations"

EMBED_LIMIT = 5  # more embeds in a row than this causes issues

USER_PREFRENCES = [
    "hemisphere",
    "timezone",
    "island",
    "friend",
    "fruit",
    "nickname",
    "creator",
]

# Based on values from datetime.isoweekday()
DAYS = {
    "monday": 1,
    "tuesday": 2,
    "wednesday": 3,
    "thursday": 4,
    "friday": 5,
    "saturday": 6,
    "sunday": 7,
}
IDAYS = dict(map(reversed, DAYS.items()))


class Validate:
    FRUITS = ["apple", "cherry", "orange", "peach", "pear"]
    HEMISPHERES = ["northern", "southern"]

    @classmethod
    def friend(cls, value):
        code = re.sub("[^0-9]", "", value)
        return code if len(code) == 12 and code.isdigit() else None

    @classmethod
    def creator(cls, value):
        code = re.sub("[^0-9]", "", value)
        return code if len(code) == 12 and code.isdigit() else None

    @classmethod
    def fruit(cls, value):
        fruit = value.lower()
        return fruit if fruit in cls.FRUITS else None

    @classmethod
    def hemisphere(cls, value):
        home = value.lower()
        return home if home in cls.HEMISPHERES else None

    @classmethod
    def nickname(cls, value):
        return value

    @classmethod
    def timezone(cls, value):
        return value if value in pytz.all_timezones_set else None

    @classmethod
    def island(cls, value):
        return value


def h(dt):
    """Convertes a datetime to something readable by a human."""
    if hasattr(dt, "tz_convert"):  # pandas-datetime-like objects
        dt = dt.to_pydatetime()
    naive_dt = dt.replace(tzinfo=None)
    return naturaltime(naive_dt)


def day_and_time(dt):
    """Converts a datetime to a day and time of day, eg: Monday pm."""
    day = IDAYS[dt.isoweekday()]
    am_pm = "am" if dt.hour < 12 else "pm"
    return f"{day.title()} {am_pm}"


def sanitize_collectable(item):
    return unidecode(item.strip().lower())


def humanize_months(row):
    """Generator that humanizes months from row data where each month is a column."""
    ABBR = {
        0: "Jan",
        1: "Feb",
        2: "Mar",
        3: "Apr",
        4: "May",
        5: "Jun",
        6: "Jul",
        7: "Aug",
        8: "Sep",
        9: "Oct",
        10: "Nov",
        11: "Dec",
    }
    months = [
        row["jan"],
        row["feb"],
        row["mar"],
        row["apr"],
        row["may"],
        row["jun"],
        row["jul"],
        row["aug"],
        row["sep"],
        row["oct"],
        row["nov"],
        row["dec"],
    ]
    start = None
    for m, inc in enumerate(months):
        if inc and start is None:
            start = m  # start of a range
        elif not inc and start is None:
            continue  # range hasn't started yet
        elif inc and start is not None:
            continue  # continuance of a range
        else:
            lhs = ABBR[start]
            rhs = ABBR[m - 1]
            if lhs != rhs:
                yield f"{lhs} - {rhs}"  # previous element ended a range
            else:
                yield f"{lhs}"  # captures a lone element
            start = None
    if start == 0:
        yield "the entire year"  # capture total range
    elif start is not None:
        lhs = ABBR[start]
        rhs = ABBR[11]
        if lhs != rhs:
            yield f"{lhs} - {rhs}"  # capture a trailing range
        else:
            yield f"{lhs}"  # captures a trailing lone element


def discord_user_from_name(channel, name):
    """Returns the discord user from the given channel and name."""
    if name is None:
        return None
    lname = name.lower()
    members = channel.members
    return next(filter(lambda member: lname in str(member).lower(), members), None)


def discord_user_from_id(channel, user_id):
    """Returns the discord user from the given channel and user id."""
    if user_id is None:
        return None
    iid = int(user_id)
    members = channel.members
    return next(filter(lambda member: iid == member.id, members), None)


def discord_user_name(channel, name_or_id):
    """Returns the discord user name from the given channel and name or id."""
    if not name_or_id:
        return None
    user = (
        discord_user_from_id(channel, name_or_id)
        if isinstance(name_or_id, int) or name_or_id.isdigit()
        else discord_user_from_name(channel, name_or_id)
    )
    return str(user) if user else None


def discord_user_id(channel, name):
    """Returns the discord user id name from the given channel and name."""
    if not name:
        return None
    return getattr(discord_user_from_name(channel, name), "id", None)


def is_turbot_admin(channel, user_or_member):
    """Checks to see if given user or member has the Turbot Admin role on this server."""
    member = (
        user_or_member
        if hasattr(user_or_member, "roles")  # members have a roles property
        else channel.guild.get_member(user_or_member.id)  # but users don't
    )
    return any(role.name == "Turbot Admin" for role in member.roles) if member else False


def command(f):
    f.is_command = True
    return f


class Turbot(discord.Client):
    """Discord turnip bot"""

    def __init__(
        self, token="", channels=[], db_dir=DEFAULT_DB_DIR, log_level=None,
    ):
        if log_level:  # pragma: no cover
            logging.basicConfig(level=log_level)
        super().__init__()
        self.token = token
        self.channels = channels
        self.base_prophet_url = "https://turnipprophet.io/?prices="
        self._last_backup_filename = None
        self.data = Data(db_dir=db_dir)

        # build a list of commands supported by this bot by fetching @command methods
        members = inspect.getmembers(self, predicate=inspect.ismethod)
        self._commands = [
            member[0]
            for member in members
            if hasattr(member[1], "is_command") and member[1].is_command
        ]

        # load static application data
        self.assets = Assets()

    def run(self):  # pragma: no cover
        super().run(self.token)

    def last_backup_filename(self):
        """Return the name of the last known backup file for prices or None if unknown."""
        return self._last_backup_filename

    def backup_prices(self, data):
        """Backs up the prices data to a datetime stamped file."""
        filename = datetime.now(pytz.utc).strftime(
            "prices-%Y-%m-%d.csv"  # TODO: configurable?
        )
        filepath = Path(self.data.file("prices")).parent / filename
        self._last_backup_filename = filepath
        data.to_csv(filepath, index=False)

    def _get_island_data(self, user):
        timeline = self.get_user_timeline(user.id)
        timeline_data = dict(
            zip(
                [
                    "Sunday_AM",
                    "Monday_AM",
                    "Monday_PM",
                    "Tuesday_AM",
                    "Tuesday_PM",
                    "Wednesday_AM",
                    "Wednesday_PM",
                    "Thursday_AM",
                    "Thursday_PM",
                    "Friday_AM",
                    "Friday_PM",
                    "Saturday_AM",
                    "Saturday_PM",
                ],
                timeline,
            )
        )
        timeline_data = {k: v for k, v in timeline_data.items() if v is not None}
        # TODO: Incorporate information about user's pattern from last week
        return {"initial_week": False, "timeline": timeline_data}

    def _get_predictive_graph(self, target_user, graphname):
        """Builds a predictive graph of a user's price data."""
        islands = {"islands": {}}
        island_data = self._get_island_data(target_user)
        if "Sunday_AM" not in island_data["timeline"]:
            return None
        islands["islands"][target_user.name] = island_data
        arch = Archipelago.load_json(json.dumps(islands))
        island = next(arch.islands)  # there should only be one island
        plot_models_range(
            island.name, list(island.model_group.models), island.previous_week, True
        )

        ax = plt.gca()
        if "!!ERROR!!" in ax.get_title():
            return None

        # fit the y-axis to the plotted data
        maximum = 0
        for line in ax.lines:
            for price in line.get_ydata():
                if price > maximum:
                    maximum = price
        for collection in ax.collections:
            for point in collection.get_offsets():
                _, y = point
                if y > maximum:
                    maximum = y
        ax.set_ylim(0, maximum + 50)
        ax.autoscale_view()
        plt.draw()

        plt.savefig(graphname)
        return plt

    def _get_historical_graph(self, channel, graphname):
        """Builds a historical graph of everyone's price data."""
        HOURS = mdates.HourLocator()
        HOURS_FMT = mdates.DateFormatter("%b %d %H:%M")
        TWELVEHOUR = mdates.HourLocator(interval=12)

        plt.figure(figsize=(10, 12), dpi=100)
        _, ax = plt.subplots()
        ax.xaxis.set_major_locator(TWELVEHOUR)
        ax.xaxis.set_major_formatter(HOURS_FMT)
        ax.xaxis.set_minor_locator(HOURS)

        ax.yaxis.set_minor_locator(matplotlib.ticker.MultipleLocator(5))

        priceList = self.data.prices
        legendElems = []

        found_at_least_one_user = False
        for user_id, df in priceList.groupby(by="author"):
            dates = []
            prices = []
            user_name = discord_user_name(channel, user_id)
            if not user_name:
                continue
            found_at_least_one_user = True
            legendElems.append(user_name)
            for _, row in df.iterrows():
                if row.kind == "sell":
                    prices.append(row.price)
                    dates.append(row.timestamp)
            if dates:
                plt.plot(dates, prices, linestyle="-", marker="o", label=user_name)
        if not found_at_least_one_user:
            return None

        plt.xticks(rotation=45, ha="right", rotation_mode="anchor")
        plt.subplots_adjust(left=0.05, bottom=0.2, right=0.85)
        plt.grid(b=True, which="major", color="#666666", linestyle="-")
        ax.yaxis.grid(b=True, which="minor", color="#555555", linestyle=":")
        plt.ylabel("Price")
        plt.xlabel("Time (UTC)")
        plt.title("Selling Prices")
        plt.legend(legendElems, loc="upper left", bbox_to_anchor=(1, 1))

        figure = plt.gcf()
        figure.set_size_inches(18, 9)

        plt.savefig(graphname, dpi=100)
        return plt

    def get_graph(self, channel, target_user, graphname):
        """Returns a graph of user data; the call site is responsible for closing."""
        if target_user:
            return self._get_predictive_graph(target_user, graphname)
        else:
            return self._get_historical_graph(channel, graphname)

    def generate_graph(self, channel, target_user, graphname):  # pragma: no cover
        """Generates a nice looking graph of user data."""
        plot = self.get_graph(channel, target_user, graphname)
        success = plot is not None
        plt.close("all")
        return success

    def append_price(self, author, kind, price, at):
        """Adds a price to the prices data file for the given author and kind."""
        at = datetime.now(pytz.utc) if not at else at
        at = at.astimezone(pytz.utc)  # always store data in UTC
        prices = self.data.prices
        row = pd.DataFrame(columns=prices.columns, data=[[author.id, kind, price, at]])
        prices = prices.append(row, ignore_index=True)
        self.data.commit(prices)

    def get_last_price(self, user_id):
        """Returns the last sell price for the given user id."""
        prices = self.data.prices
        last = (
            prices[(prices.author == user_id) & (prices.kind == "sell")]
            .sort_values(by=["timestamp"])
            .tail(1)
            .price
        )
        return last.iloc[0] if last.any() else None

    def get_user_prefs(self, user_id):
        users = self.data.users
        row = users[users.author == user_id].tail(1)
        if row.empty:
            return {}

        prefs = {}
        data = row.to_dict(orient="records")[0]
        for column in users.columns[1:]:
            if not data[column]:
                continue
            if column == "timezone":
                butt = data[column]
                prefs[column] = pytz.timezone(butt)
            else:
                prefs[column] = data[column]
        return prefs

    def get_user_timeline(self, user_id):
        prices = self.data.prices
        past = datetime.now(pytz.utc) - timedelta(days=12)
        yours = prices[(prices.author == user_id) & (prices.timestamp > past)]
        yours = yours.sort_values(by=["timestamp"])

        # convert all timestamps to the target user's timezone
        target_timezone = self.get_user_prefs(user_id).get("timezone", pytz.UTC)
        yours["timestamp"] = yours.timestamp.dt.tz_convert(target_timezone)

        recent_buy = yours[yours.kind == "buy"].tail(1)
        if recent_buy.empty:  # no buy found
            return [None] * 13

        buy_date = recent_buy.timestamp.iloc[0]
        if (
            buy_date.to_pydatetime().isoweekday() != DAYS["sunday"]
        ):  # buy isn't on a sunday
            return [None] * 13

        buy_price = int(recent_buy.price.iloc[0])
        sells = yours[(yours.kind == "sell") & (yours.timestamp > buy_date)]
        if sells.empty:  # no sells found after buy date
            return [buy_price] + [None] * 12

        sell_data = [
            [None, None],  # monday
            [None, None],  # tuesday
            [None, None],  # wednesday
            [None, None],  # friday
            [None, None],  # thrusday
            [None, None],  # saturday
        ]

        groups = sells.set_index("timestamp").groupby(pd.Grouper(freq="D"))
        for day, df in groups:
            day_of_week = day.to_pydatetime().isoweekday()
            if day_of_week == DAYS["sunday"]:  # no sells allowed on sundays
                continue
            for ts, row in df.iterrows():
                if ts.hour < 12:  # am
                    sell_data[day_of_week - 1][0] = int(row["price"])
                else:  # pm
                    sell_data[day_of_week - 1][1] = int(row["price"])

        timeline = [buy_price]
        for day in sell_data:
            timeline.extend(day)
        return timeline

    def to_usertime(self, author_id, dt):
        user_timezone = self.get_user_prefs(author_id).get("timezone", pytz.UTC)
        if hasattr(dt, "tz_convert"):  # pandas-datetime-like objects
            return dt.tz_convert(user_timezone)
        elif hasattr(dt, "astimezone"):  # python-datetime-like objects
            return dt.astimezone(user_timezone)
        else:  # pragma: no cover
            logging.warning(f"can't convert tz on {dt} for user {author_id}")
            return dt

    def set_user_pref(self, author, pref, value):
        users = self.data.users
        row = users[users.author == author.id].tail(1)
        if row.empty:
            users = users.append({"author": author.id, pref: value}, ignore_index=True)
        else:
            users.at[row.index, pref] = value
        self.data.commit(users)

    def paginate(self, text):
        """Discord responses must be 2000 characters of less; paginate breaks them up."""
        breakpoints = ["\n", ".", ",", "-"]
        remaining = text
        while len(remaining) > 2000:
            breakpoint = 1999

            for char in breakpoints:
                index = remaining.rfind(char, 1800, 1999)
                if index != -1:
                    breakpoint = index
                    break

            message = remaining[0 : breakpoint + 1]
            yield message
            remaining = remaining[breakpoint + 1 :]
            last_line_end = message.rfind("\n")
            if last_line_end != -1 and len(message) > last_line_end + 1:
                last_line_start = last_line_end + 1
            else:
                last_line_start = 0
            if message[last_line_start] == ">":
                remaining = f"> {remaining}"

        yield remaining

    @property
    def commands(self):
        """Returns a list of commands supported by this bot."""
        return self._commands

    async def process(self, message):
        """Process a command message."""
        tokens = message.content.split(" ")
        request, params = tokens[0].lstrip("!").lower(), tokens[1:]
        params = list(filter(None, params))  # ignore any empty string parameters
        if not request:
            return
        matching = [command for command in self.commands if command.startswith(request)]
        if not matching:
            await message.channel.send(s("not_a_command", request=request), file=None)
            return
        if len(matching) > 1 and request not in matching:
            possible = ", ".join(f"!{m}" for m in matching)
            await message.channel.send(s("did_you_mean", possible=possible), file=None)
        else:
            command = request if request in matching else matching[0]
            logging.debug("%s (author=%s, params=%s)", command, message.author, params)
            method = getattr(self, command)
            async with message.channel.typing():
                response, attachment = method(message.channel, message.author, params)
            if not isinstance(response, list):
                response = [response]
            last_reply_index = len(response) - 1
            for n, reply in enumerate(response):
                if isinstance(reply, str):
                    pages = list(self.paginate(reply))
                    last_page_index = len(pages) - 1
                    for i, page in enumerate(pages):
                        file = (
                            attachment
                            if attachment is not None
                            and i == last_page_index
                            and n == last_reply_index
                            else None
                        )
                        await message.channel.send(page, file=file)
                elif isinstance(reply, discord.embeds.Embed):
                    file = (
                        attachment
                        if attachment is not None and n == last_reply_index
                        else None
                    )
                    await message.channel.send(embed=reply, file=file)
                else:
                    raise RuntimeError("non-string non-embed reply not supported")

    ##############################
    # Discord Client Behavior
    ##############################

    async def on_message(self, message):
        """Behavior when the client gets a message from Discord."""
        if (
            str(message.channel.type) == "text"
            and message.author.id != self.user.id
            and message.channel.name in self.channels
            and message.content.startswith("!")
        ):
            await self.process(message)

    async def on_ready(self):
        """Behavior when the client has successfully connected to Discord."""
        logging.debug("logged in as %s", self.user)

    ##############################
    # Bot Command Functions
    ##############################

    # Any method of this class with a name that is decorated by @command is detected as a
    # bot command. These methods should have a signature like:
    #
    #     @command
    #     def command_name(self, channel, author, params)
    #
    # - `channel` is the Discord channel where the command message was sent.
    # - `author` is the Discord author who sent the command.
    # - `params` are any space delimitered parameters also sent with the command.
    #
    # The return value for a command method can be `(string, discord.File)` where the
    # string is the response message the bot should send to the channel and the file
    # object is an attachment to send with the message. For no attachment, use `None`.
    #
    # You can also return `(embed, discord.File)` to respond with an embed (plus the
    # optional attachment as well if so desired).
    #
    # And finally you can also respond with a list if you want the bot to make multiple
    # replies. This works with both embeds and strings. For example:
    #
    #     return ["send", "multiple", "replies"], None
    #
    # Would trigger the bot to send three messages to the channel with no attachment.
    #
    # The docstring used for the command method will be automatically used as the help
    # message for the command. To document commands with parameters use a @ to delimit
    # the help message from the parameter documentation. For example:
    #
    #     """This is the help message for your command. @ [and] [these] [are] [params]"""
    #
    # A [parameter] is optional whereas a <parameter> is required.

    @command
    def help(self, channel, author, params):
        """
        Shows this help screen.
        """
        usage = "__**Turbot Help!**__\n"
        for command in self.commands:
            method = getattr(self, command)
            doc = method.__doc__.split("@")
            use, params = doc[0], ", ".join([param.strip() for param in doc[1:]])
            use = inspect.cleandoc(use)
            use = use.replace("\n", " ")

            title = f"**!{command}**"
            if params:
                title = f"{title} _{params}_"
            usage += f"\n{title}"
            usage += f"\n>  {use}"
            usage += "\n"
        usage += "\n_Turbot created by TheAstropath_"
        return usage, None

    class _PriceTimeError(Exception):
        def __init__(self, key):
            self.key = key

    def _get_price_time(self, user_id, params):
        if len(params) == 1:
            return None

        if len(params) != 3:
            raise Turbot._PriceTimeError("price_time_invalid")

        day_of_week = params[1].lower()
        if day_of_week not in DAYS:
            raise Turbot._PriceTimeError("day_of_week_invalid")

        time_of_day = params[2].lower()
        if time_of_day not in ["morning", "evening"]:
            raise Turbot._PriceTimeError("time_of_day_invalid")

        now = self.to_usertime(user_id, datetime.now(pytz.utc))
        start = now - timedelta(days=now.isoweekday() % 7)  # start of week
        start = datetime(start.year, start.month, start.day, tzinfo=start.tzinfo)
        day_offset = DAYS[day_of_week] % 7
        hour_offset = 13 if time_of_day == "evening" else 0
        return start + timedelta(days=day_offset, hours=hour_offset)

    @command
    def sell(self, channel, author, params):
        """
        Log the price that you can sell turnips for on your island.
        @ <price> [day time-of-day]
        """
        if not params:
            return s("sell_no_params"), None

        price = params[0]
        if not price.isnumeric():
            return s("sell_nonnumeric_price"), None

        price = int(price)
        if price <= 0:
            return s("sell_nonpositive_price"), None

        last_price = self.get_last_price(author.id)

        try:
            price_time = self._get_price_time(author.id, params)
        except Turbot._PriceTimeError as err:
            return s(err.key), None

        logging.debug("saving sell price of %s bells for user id %s", price, author.id)
        self.append_price(author=author, kind="sell", price=price, at=price_time)

        key = (
            "sell_new_price"
            if not last_price or price_time is not None
            else "sell_higher_price"
            if price > last_price
            else "sell_lower_price"
            if price < last_price
            else "sell_same_price"
        )
        return s(key, price=price, name=author, last_price=last_price), None

    @command
    def buy(self, channel, author, params):
        """
        Log the price that you can buy turnips from Daisy Mae on your island.
        @ <price> [day time]
        """
        if not params:
            return s("buy_no_params"), None

        price = params[0]
        if not price.isnumeric():
            return s("buy_nonnumeric_price"), None

        price = int(price)
        if price <= 0:
            return s("buy_nonpositive_price"), None

        try:
            price_time = self._get_price_time(author.id, params)
        except Turbot._PriceTimeError as err:
            return s(err.key), None

        logging.debug("saving buy price of %s bells for user id %s", price, author.id)
        self.append_price(author=author, kind="buy", price=price, at=price_time)

        return s("buy", price=price, name=author), None

    @command
    def reset(self, channel, author, params):
        """
        Only Turbot Admin members can run this command. Generates a final graph for use
        with !lastweek and resets all data for all users.
        """
        if not is_turbot_admin(channel, author):
            return s("not_admin"), None

        self.generate_graph(channel, None, LASTWEEKCMD_FILE)
        prices = self.data.prices
        self.backup_prices(prices)

        buys = prices[prices.kind == "buy"].sort_values(by="timestamp")
        idx = buys.groupby(by="author")["timestamp"].idxmax()
        prices = buys.loc[idx]
        self.data.commit(prices)
        return s("reset"), None

    @command
    def lastweek(self, channel, author, params):
        """
        Displays the final graph from the last week before the data was reset.
        """
        if not Path(LASTWEEKCMD_FILE).exists():
            return s("lastweek_none"), None
        return s("lastweek"), discord.File(LASTWEEKCMD_FILE)

    @command
    def graph(self, channel, author, params):
        """
        Generates a historical graph of turnip prices for all users.
        """
        success = self.generate_graph(channel, None, GRAPHCMD_FILE)
        attach = discord.File(GRAPHCMD_FILE) if success else None
        return s("graph_all_users"), attach

    @command
    def history(self, channel, author, params):
        """
        Show the historical turnip prices for a user. If no user is specified, it will
        display your own prices. @ [user]
        """
        target = author.id if not params else params[0]
        target_name = discord_user_name(channel, target)
        target_id = discord_user_id(channel, target_name)
        if not target_name or not target_id:
            return s("cant_find_user", name=target), None

        prices = self.data.prices
        yours = prices[prices.author == target_id]
        lines = [s("history_header", name=target_name)]
        for _, row in yours.iterrows():
            time = self.to_usertime(target_id, row.timestamp)
            lines.append(
                s(
                    f"history_{row.kind}",
                    price=row.price,
                    timestamp=h(time),
                    day_and_time=day_and_time(time),
                )
            )
        return "\n".join(lines), None

    @command
    def oops(self, channel, author, params):
        """
        Remove your last logged turnip price.
        """
        target = author.id
        target_name = discord_user_name(channel, target)
        prices = self.data.prices
        prices = prices.drop(prices[prices.author == author.id].tail(1).index)
        self.data.commit(prices)
        return s("oops", name=target_name), None

    @command
    def clear(self, channel, author, params):
        """
        Clears all of your own historical turnip prices.
        """
        user_id = discord_user_id(channel, str(author))
        prices = self.data.prices
        prices = prices[prices.author != user_id]
        self.data.commit(prices)
        return s("clear", name=author), None

    @command
    def best(self, channel, author, params):
        """
        Finds the best, most recent, buy or sell price currently available.
        The default is to look for the best sell.
        @ [buy|sell]
        """
        kind = params[0].lower() if params else "sell"
        if kind not in ["sell", "buy"]:
            return s("best_invalid_param"), None

        prices = self.data.prices
        past = datetime.now(pytz.utc) - timedelta(hours=12)
        sells = prices[(prices.kind == kind) & (prices.timestamp > past)]
        idx = sells.groupby(by="author").price.transform(max) == sells.price
        bests = sells[idx].sort_values(by="price", ascending=kind == "buy")
        lines = [s(f"best_{kind}_header")]
        for _, row in bests.iterrows():
            name = discord_user_from_id(channel, row.author)
            timestamp = h(self.to_usertime(row.author, row.timestamp))
            lines.append(s("best", name=name, price=row.price, timestamp=timestamp))
        return "\n".join(lines), None

    @command
    def collect(self, channel, author, params):
        """
        Mark collectables as donated to your museum. The names must match the in-game item
        name exactly. @ <comma, separated, list, of, things>
        """
        if not params:
            return s("collect_no_params"), None

        items = set(sanitize_collectable(item) for item in " ".join(params).split(","))
        valid, invalid = self.assets.validate(items)

        lines = []

        def add_lines(kind, valid_items, fullset):
            if not valid_items:
                return
            store = getattr(self.data, kind)
            yours = store[store.author == author.id]
            dupes = yours.loc[yours.name.isin(valid_items)].name.values.tolist()
            new_items = list(set(valid_items) - set(dupes))
            new_data = [[author.id, name] for name in new_items]
            new_fossils = pd.DataFrame(columns=store.columns, data=new_data)
            store = store.append(new_fossils, ignore_index=True)
            yours = store[store.author == author.id]  # re-fetch for congrats
            self.data.commit(store)
            if new_items:
                lines.append(s(f"collect_{kind}_new", items=", ".join(sorted(new_items))))
            if dupes:
                lines.append(s(f"collect_{kind}_dupe", items=", ".join(sorted(dupes))))
            if len(fullset) == len(yours.index):
                lines.append(s(f"congrats_all_{kind}"))

        for kind in self.assets.collectables:
            add_lines(kind, valid[kind], self.assets[kind].all)

        if invalid:
            lines.append(s("invalid_collectable", items=", ".join(sorted(invalid))))

        return "\n".join(lines), None

    @command
    def uncollect(self, channel, author, params):
        """
        Unmark collectables as donated to your museum. The names must match the in-game
        item name exactly. @ <comma, separated, list, of, things>
        """
        if not params:
            return s("uncollect_no_params"), None

        items = set(sanitize_collectable(item) for item in " ".join(params).split(","))
        valid, invalid = self.assets.validate(items)

        lines = []

        def add_lines(kind, valid_items):
            store = getattr(self.data, kind)
            yours = store[store.author == author.id]
            previously_collected = yours.loc[yours.name.isin(valid_items)]
            deleted = set(previously_collected.name.values.tolist())
            didnt_have = valid_items - deleted
            store = store.drop(previously_collected.index)
            self.data.commit(store)
            if deleted:
                items_str = ", ".join(sorted(deleted))
                lines.append(s(f"uncollect_{kind}_deleted", items=items_str))
            if didnt_have:
                items_str = ", ".join(sorted(didnt_have))
                lines.append(s(f"uncollect_{kind}_already", items=items_str))

        for kind in self.assets.collectables:
            add_lines(kind, valid[kind])

        if invalid:
            lines.append(s("invalid_collectable", items=", ".join(sorted(invalid))))

        return "\n".join(lines), None

    @command
    def search(self, channel, author, params):
        """
        Searches all users to see who needs the given collectables. The names must match
        the in-game item name, and more than one can be provided if separated by commas.
        @ <comma, separated, list, of, collectables>
        """
        if not params:
            return s("search_no_params"), None

        items = set(sanitize_collectable(item) for item in " ".join(params).split(","))
        valid, invalid = self.assets.validate(items)

        def get_results(kind, valid_items):
            store = getattr(self.data, kind)
            users = store.author.unique()
            results = defaultdict(list)
            for collected_item in valid_items:
                havers = store[store.name == collected_item].author.unique()
                needers = np.setdiff1d(users, havers).tolist()
                for needer in needers:
                    name = discord_user_from_id(channel, needer)
                    results[name].append(collected_item)
            return results

        results = {}
        for kind in self.assets.collectables:
            results[kind] = get_results(kind, valid[kind])

        if not any(result for result in results.values()) and not invalid:
            return s("search_all_not_needed"), None

        searched = set()
        for valid_items in valid.values():
            searched.update(valid_items)

        needed = set()
        for result in results.values():
            for items in result.values():
                needed.update(items)

        not_needed = searched - needed

        lines = []
        for kind, result in results.items():
            for name, items in result.items():
                items_str = ", ".join(sorted(items))
                lines.append(s(f"search_{kind}_row", name=name, items=items_str))
        if not_needed:
            items_str = ", ".join(sorted(not_needed))
            lines.append(s("search_not_needed", items=items_str))
        if invalid:
            lines.append(s("search_invalid", items=", ".join(sorted(invalid))))
        return "\n".join(sorted(lines)), None

    @command
    def uncollected(self, channel, author, params):
        """
        Lists all collectables that you still need to donate. If a user is provided, it
        gives the same information for that user instead. @ [user]
        """
        target = author.id if not params else params[0]
        target_name = discord_user_name(channel, target)
        target_id = discord_user_id(channel, target_name)
        if not target_name or not target_id:
            return s("cant_find_user", name=target), None

        lines = []

        def add_lines(kind, fullset):
            store = getattr(self.data, kind)
            your_items = store[store.author == target_id]
            collected_items = set(your_items.name.unique())
            remaining_items = fullset - collected_items
            if remaining_items:
                count_key = f"uncollected_{kind}_count"
                remaining_key = f"uncollected_{kind}_remaining"
                count = len(remaining_items)
                items_str = ", ".join(sorted(remaining_items))
                lines.append(s(count_key, count=count, name=target_name))
                lines.append(s(remaining_key, items=items_str))
            else:
                lines.append(s(f"congrats_all_{kind}"))

        for kind in self.assets.collectables:
            add_lines(kind, self.assets[kind].all)

        return "\n".join(lines), None

    @command
    def needed(self, channel, author, params):
        """
        Lists all the needed items for all the channel members. As the only parameter
        give the name of the kind of collectable to return.
        @ <fossils|bugs|fish|art|songs>
        """
        if not params:
            return s("needed_no_param"), None

        kind = params[0].lower()
        if kind not in ["fossils", "bugs", "fish", "art", "songs"]:
            return s("needed_invalid_param"), None

        fullset = self.assets[kind].all
        store = getattr(self.data, kind)
        authors = [member.id for member in channel.members if member.id != self.user.id]
        total = pd.DataFrame(list(product(authors, fullset)), columns=store.columns)
        merged = total.merge(store, indicator=True, how="outer")
        needed = merged[merged["_merge"] == "left_only"]

        limit = 10
        lines = []

        for user, df in needed.groupby(by="author"):
            name = discord_user_name(channel, user)
            items_list = sorted([row["name"] for _, row in df.iterrows()])
            if len(items_list) == len(fullset):
                continue
            elif len(items_list) > limit:
                items_str = s("needed_lots", limit=limit, kind=kind)
            else:
                items_str = ", ".join(items_list)
            lines.append(s("needed", name=name, items=items_str))

        if not lines:
            return s("needed_none", kind=kind), None

        return "\n".join(sorted(lines)), None

    @command
    def collected(self, channel, author, params):
        """
        Lists all collectables that you have already donated. If a user is provided, it
        gives the same information for that user instead. @ [user]
        """
        target = author.id if not params else params[0]
        target_name = discord_user_name(channel, target)
        target_id = discord_user_id(channel, target_name)
        if not target_name or not target_id:
            return s("cant_find_user", name=target), None

        def get_collection(kind):
            store = getattr(self.data, kind)
            your_items = store[store.author == target_id]
            return set(your_items.name.unique())

        collected_items = {
            kind: get_collection(kind) for kind in self.assets.collectables
        }

        all_items = {
            kind: len(collected_items[kind]) == len(self.assets[kind].all)
            for kind in self.assets.collectables
        }

        lines = []

        if any(flag for flag in all_items.values()):
            for kind, flag in all_items.items():
                if flag:
                    lines.append(s(f"congrats_all_{kind}"))
            if all(flag for flag in all_items.values()):
                return "\n".join(lines), None

        for kind, items in collected_items.items():
            if items and not all_items[kind]:
                key = f"collected_{kind}"
                count = len(items)
                items_str = ", ".join(sorted(items))
                lines.append(s(key, name=target_name, count=count, items=items_str))

        return "\n".join(lines), None

    @command
    def predict(self, channel, author, params):
        """
        Get a link to a prediction calculator for a price history. @ [user]
        """
        target = author.id if not params else params[0]
        target_name = discord_user_name(channel, target)
        target_id = discord_user_id(channel, target_name)
        target_user = discord_user_from_id(channel, target_id)
        if not target_name or not target_id:
            return s("cant_find_user", name=target), None

        timeline = self.get_user_timeline(target_id)
        if not timeline[0]:
            return s("cant_find_buy", name=target_name), None

        success = self.generate_graph(channel, target_user, GRAPHCMD_FILE)
        query = ".".join((str(price) if price else "") for price in timeline).rstrip(".")
        url = f"{self.base_prophet_url}{query}"
        attach = discord.File(GRAPHCMD_FILE) if success else None
        return s("predict", name=target_name, url=url), attach

    @command
    def pref(self, channel, author, params):
        """
        Set one of your user preferences. @ <preference> <value>
        """
        if not params:
            return s("pref_no_params", prefs=", ".join(USER_PREFRENCES)), None

        pref = params[0].lower()
        if pref not in USER_PREFRENCES:
            return s("pref_invalid_pref", prefs=", ".join(USER_PREFRENCES)), None
        if len(params) <= 1:
            return s("pref_no_value", pref=pref), None

        value = " ".join(params[1:])
        validated_value = getattr(Validate, pref)(value)
        if not validated_value:
            return s(f"{pref}_invalid"), None

        self.set_user_pref(author, pref, validated_value)
        return s("pref", pref=pref, name=author), None

    @command
    def count(self, channel, author, params):
        """
        Provides a count of the number of pieces of collectables for the comma-separated
        list of users. @ [comma, separated, list, of, users]
        """
        if not params:
            user = discord_user_from_id(channel, author.id)
            valid = [(user.name, user.id)]
            invalid = []
        else:
            users = set(sanitize_collectable(p) for p in " ".join(params).split(","))
            valid = []
            invalid = []
            for user in users:
                user_name = discord_user_name(channel, user)
                user_id = discord_user_id(channel, user_name)
                if user_name and user_id:
                    valid.append((user_name, user_id))
                else:
                    invalid.append(user)

        lines = []

        def add_valid_lines(kind, fullset):
            lines.append(s(f"count_{kind}_valid_header"))
            store = getattr(self.data, kind)
            for user_name, user_id in sorted(valid):
                yours = store[store.author == user_id]
                collected = set(yours.name.unique())
                remaining = fullset - collected
                count = len(remaining)
                lines.append(s(f"count_{kind}_valid", name=user_name, count=count))

        if valid:
            for kind in self.assets.collectables:
                add_valid_lines(kind, self.assets[kind].all)

        if invalid:
            lines.append(s("count_invalid_header"))
            for user in invalid:
                lines.append(s("count_invalid", name=user))

        return "\n".join(lines), None

    @command
    def art(self, channel, author, params):
        """
        Get info about pieces of art that are available
        @ [comma, separated, list, of, art, pieces]
        """
        response = ""
        if params:
            items = set(
                sanitize_collectable(item) for item in " ".join(params).split(",")
            )
            validset, invalidset = self.assets.validate(items, kinds=["art"])
            art_data = self.assets["art"].data
            valid = sorted(list(validset["art"]))
            invalid = sorted(list(invalidset))
            lines = []
            response = s("art_header") + "\n"
            for art in valid:
                piece = art_data[art_data.name == art].iloc[0]
                if piece["has_fake"]:
                    lines.append(
                        s(
                            "art_fake",
                            name=piece["name"].title(),
                            desc=piece["fake_description"],
                            real_url=piece["real_image_url"],
                            fake_url=piece["fake_image_url"],
                        )
                    )
                else:
                    lines.append(
                        s(
                            "art_real",
                            name=piece["name"].title(),
                            real_url=piece["real_image_url"],
                        )
                    )

            response += "\n> \n".join(lines)

            if invalid:
                response += "\n" + (s("art_invalid", items=", ".join(invalid)))

        else:
            items = self.assets["art"].all
            response = s("allart", list=", ".join(sorted(items)))

        return response, None

    def creatures_available_now(self, now, df):
        """Returns the names of creatures in df that are available right now."""
        for _, row in df.iterrows():
            time = row["time"]
            if time.lower() == "all day":
                yield row["name"]
            else:
                ranges = time.lower().split("&")
                for r in ranges:
                    lhs, rhs = [t.strip().lower() for t in r.split("-")]
                    lhs_hour, lhs_ampm = lhs.split(" ")
                    rhs_hour, rhs_ampm = rhs.split(" ")
                    lhs_hour, rhs_hour = int(lhs_hour), int(rhs_hour)
                    now_hour = now.hour
                    now_ampm = "am" if now_hour < 12 else "pm"

                    # | lhs_ampm | rhs_ampm | now_ampm | yield if (use _hour vars) |
                    # | -------- | -------- | -------- | ------------------------- |
                    # | am       | am       | am       | lhs <= now <= rhs         |
                    # | am       | am       | pm       | False                     |
                    # | am       | pm       | am       | lhs <= now <= rhs + 12    |
                    # | am       | pm       | pm       | lhs <= now <= rhs + 12    |
                    # | pm       | am       | am       | now <= rhs                |
                    # | pm       | am       | pm       | now >= lhs                |
                    # | pm       | pm       | am       | False                     |
                    # | pm       | pm       | pm       | lhs <= now <= rhs         |
                    if lhs_ampm == "am" and rhs_ampm == "am" and now_ampm == "am":
                        if lhs_hour <= now_hour <= rhs_hour:
                            yield row["name"]
                    elif lhs_ampm == "am" and rhs_ampm == "am" and now_ampm == "pm":
                        continue
                    elif lhs_ampm == "am" and rhs_ampm == "pm" and now_ampm == "am":
                        if lhs_hour <= now_hour <= rhs_hour + 12:
                            yield row["name"]
                    elif lhs_ampm == "am" and rhs_ampm == "pm" and now_ampm == "pm":
                        if lhs_hour <= now_hour <= rhs_hour + 12:
                            yield row["name"]
                    elif lhs_ampm == "pm" and rhs_ampm == "am" and now_ampm == "am":
                        if now_hour <= rhs_hour:
                            yield row["name"]
                    elif lhs_ampm == "pm" and rhs_ampm == "am" and now_ampm == "pm":
                        if now_hour >= lhs_hour:
                            yield row["name"]
                    elif lhs_ampm == "pm" and rhs_ampm == "pm" and now_ampm == "am":
                        continue
                    else:  # lhs_ampm == "pm" and rhs_ampm == "pm" and now_ampm == "pm"
                        if lhs_hour <= now_hour <= rhs_hour:
                            yield row["name"]

    def _creatures(self, *, author, params, kind, source, force_text=False):
        """The fish and bugs commands are so similar; I factored them out to a helper."""
        hemisphere = self.get_user_prefs(author.id).get("hemisphere", None)
        if not hemisphere:
            return s("no_hemisphere")

        now = self.to_usertime(author.id, datetime.now(pytz.utc))
        this_month = now.strftime("%b").lower()
        first_this_month = now.replace(day=1)
        last_month = (first_this_month - timedelta(days=1)).strftime("%b").lower()
        next_month = (first_this_month + relativedelta(months=1)).strftime("%b").lower()
        available = source[(source.hemisphere == hemisphere) & (source[this_month] == 1)]

        def details(row):
            alert = (
                "**GONE NEXT MONTH!**"
                if not row[next_month]
                else "_New this month_"
                if not row[last_month]
                else ""
            )
            months = ", ".join(list(humanize_months(row)))
            return {
                **row,
                "name": row["name"].capitalize(),
                "months": months,
                "alert": alert,
            }

        if params:
            user_input = " ".join(params)
            search = user_input.lower()
            if search == "leaving":
                found = available[available[next_month] == 0]
            elif search == "arriving":
                found = available[available[last_month] == 0]
            else:
                found = available[available.name.str.contains(search)]

            if found.empty:
                return s(f"{kind}_none_found", search=user_input)
        else:
            if kind == "fish":
                caught = self.data.fish
                caught = caught[caught.author == author.id]
            else:  # kind == "bugs"
                caught = self.data.bugs
                caught = caught[caught.author == author.id]

            if caught is not None and not caught.empty:
                already = available.loc[available.name.isin(caught.name)]
                available = available.drop(already.index)

            found = available

            if found.empty:
                return s(f"{kind}_none_available")

        now_names = set(self.creatures_available_now(now, found))
        found_now = found[found.name.isin(now_names)]
        found_this_month = found[~found.name.isin(now_names)]

        def get_response(df, force_text):
            if force_text or len(df) > EMBED_LIMIT:
                lines = [s(kind, **details(row)) for _, row in df.iterrows()]
                return "\n".join(sorted(lines))

            response = []
            rows = [row for _, row in df.iterrows()]
            for row in sorted(rows, key=lambda r: r["name"]):
                info = details(row)
                embed = discord.Embed(title=info["name"])
                embed.set_thumbnail(url=info["image"])
                embed.add_field(name="price", value=info["price"])
                embed.add_field(name="location", value=info["location"])
                if "shadow" in info:
                    embed.add_field(name="shadow size", value=info["shadow"])
                embed.add_field(name="available", value=info["time"])
                embed.add_field(name="during", value=info["months"])
                if info["alert"]:
                    embed.add_field(name="alert", value=info["alert"])
                    if "GONE" in info["alert"]:
                        embed.color = discord.Color.orange()
                    else:
                        embed.color = discord.Color.blue()
                response.append(embed)
            return response

        response_now = get_response(found_now, force_text=force_text)
        response_this_month = get_response(found_this_month, force_text=True)

        responses = []

        if response_now:
            header = s("creature_available_now", kind=kind.title())
            if isinstance(response_now, str):
                responses.append(f"{header}\n{response_now}")
            else:
                responses.append(header)
                responses.extend(response_now)

        if response_this_month:
            # response_this_month is always forced to be text
            header = s("creature_available_this_month", kind=kind.title())
            responses.append(f"{header}\n{response_this_month}")

        def add_header(responses):
            if kind != "bugs":
                return responses
            if random.randint(0, 100) > 70:
                responses.insert(0, s("bugs_header"))
            return responses

        return add_header(responses)

    @command
    def fish(self, channel, author, params):
        """
        Tells you what fish are available now in your hemisphere.
        @ [name|leaving|arriving]
        """
        source = self.assets["fish"].data
        found = self._creatures(author=author, params=params, kind="fish", source=source)
        return found, None

    @command
    def bugs(self, channel, author, params):
        """
        Tells you what bugs are available now in your hemisphere.
        @ [name|leaving|arriving]
        """
        source = self.assets["bugs"].data
        found = self._creatures(author=author, params=params, kind="bugs", source=source)
        return found, None

    @command
    def new(self, channel, author, params):
        """
        Tells you what new things available in your hemisphere right now.
        """
        bugs_data = self.assets["bugs"].data
        bugs = self._creatures(
            author=author,
            params=["arriving"],
            kind="bugs",
            source=bugs_data,
            force_text=True,
        )
        fish_data = self.assets["fish"].data
        fish = self._creatures(
            author=author,
            params=["arriving"],
            kind="fish",
            source=fish_data,
            force_text=True,
        )
        return [*bugs, *fish], None

    def _info_embed(self, user):
        prefs = self.get_user_prefs(user.id)

        embed = discord.Embed(title=user.name)
        embed.set_thumbnail(url=user.avatar_url)

        nickname = prefs.get("nickname", None)
        if nickname:
            embed.add_field(name="Nickname", value=nickname)
        else:
            embed.add_field(name="Nickname", value="Not set")

        code = prefs.get("friend", None)
        if code:
            code_str = f"SW-{code[0:4]}-{code[4:8]}-{code[8:12]}"
            embed.add_field(name="Friend code", value=code_str)
        else:
            embed.add_field(name="Friend code", value="Not set")

        code = prefs.get("creator", None)
        if code:
            code_str = f"MA-{code[0:4]}-{code[4:8]}-{code[8:12]}"
            embed.add_field(name="Creator code", value=code_str)
        else:
            embed.add_field(name="Creator code", value="Not set")

        island = prefs.get("island", "Not set")
        embed.add_field(name="Island", value=island)

        hemisphere = prefs.get("hemisphere", "Not set").title()
        embed.add_field(name="Hemisphere", value=hemisphere)

        fruit = prefs.get("fruit", "Not set").title()
        embed.add_field(name="Native fruit", value=fruit)

        now = self.to_usertime(user.id, datetime.now(pytz.UTC))
        current_time = now.strftime("%I:%M %p %Z")
        embed.set_footer(text=f"Current time is {current_time}")

        return embed

    @command
    def info(self, channel, author, params):
        """
        Gives you information on a user. @ [user]
        """
        if len(params) < 1:
            return s("info_no_params"), None

        query = " ".join(params).lower()  # allow spaces in names

        for _, row in self.data.users.iterrows():
            user_id = int(row["author"])
            user_name = discord_user_name(channel, user_id)
            if not user_name:
                continue
            if user_name.lower().find(query) != -1:
                user = discord_user_from_name(channel, user_name)
                return self._info_embed(user), None

        # check if user exists, they just don't have any info
        for member in channel.members:
            if member.name.lower().find(query) != -1:
                return s("info_no_prefs", user=member), None

        return s("info_not_found"), None

    @command
    def about(self, channel, author, params):
        """
        Get information about Turbot.
        """
        embed = discord.Embed(title="Turbot")
        embed.set_thumbnail(
            url="https://raw.githubusercontent.com/theastropath/turbot/master/turbot.png"
        )
        embed.add_field(
            name="Version",
            value=f"[{__version__}](https://pypi.org/project/turbot/{__version__}/)",
        )
        embed.add_field(name="Package", value="[PyPI](https://pypi.org/project/turbot/)")
        embed.add_field(
            name="Author", value="[TheAstropath](https://github.com/theastropath)"
        )
        embed.add_field(
            name="Maintainer", value="[lexicalunit](https://github.com/lexicaluit)"
        )
        embed.description = (
            "A Discord bot for everything _Animal Crossing: New Horizons._\n"
            "\n"
            "Use the command `!help` for usage details. Having issues with Turbot? "
            "Please [report bugs](https://github.com/theastropath/turbot/issues)!\n"
        )
        embed.url = "https://github.com/theastropath/turbot"
        embed.set_footer(text="MIT © TheAstropath, lexicalunit et al")
        embed.color = discord.Color(0xFFFDC3)
        return embed, None


def get_token(token_file):  # pragma: no cover
    """Returns the discord token from the environment or your token config file."""
    token = getenv("TURBOT_TOKEN", None)
    if token:
        return token

    try:
        with open(token_file, "r") as f:
            return f.readline().strip()
    except IOError as e:
        with redirect_stdout(sys.stderr):
            print("error:", e)
            print(f"put your discord token in a file named '{token_file}'")
        sys.exit(1)


def get_channels(channels_file):  # pragma: no cover
    """Returns the authorized channels the environment or your channels config file."""
    channels = getenv("TURBOT_CHANNELS", None)
    if channels:
        return channels.split(";")

    try:
        with open(channels_file, "r") as channels_file:
            return [line.strip() for line in channels_file.readlines()]
    except IOError:
        return []


def apply_migrations():  # pragma: no cover
    """Applies any migration scripts that haven't already been applied."""

    def migration_key(script):
        return int(script.name[0:3])

    applied = set()
    if MIGRATIONS_FILE.exists():
        with open(MIGRATIONS_FILE) as f:
            applied.update(line.strip() for line in f.readlines())
    with open(MIGRATIONS_FILE, "a") as f:
        for migration in sorted(MIGRATIONS_DIR.glob("*.py"), key=migration_key):
            if migration.name not in applied:
                print(f"applying migration {migration.name}...", end=" ")
                cmd = ["python3", migration]
                proc = run(cmd, capture_output=True)
                if proc.returncode != 0:
                    print()
                    print(proc.stdout.decode("utf-8"))
                    print(proc.stderr.decode("utf-8"), file=sys.stderr)
                    sys.exit(1)
                print("done")
                f.write(f"{migration.name}\n")


@click.command()
@click.option(
    "-l",
    "--log-level",
    type=click.Choice(["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG"]),
    default="ERROR",
)
@click.option("-v", "--verbose", count=True, help="sets log level to DEBUG")
@click.option(
    "-b",
    "--bot-token-file",
    default=DEFAULT_CONFIG_TOKEN,
    help="read your discord bot token from this file; "
    "you can also set the token directly via the environment variable TURBOT_TOKEN",
)
@click.option(
    "-c",
    "--channel",
    multiple=True,
    help=(
        "authorize a channel; use this multiple times to authorize multiple channels; "
        "you can also set the list of channels via the TURBOT_CHANNELS environment "
        "variable separated by ; (semicolon)."
    ),
)
@click.option(
    "-a",
    "--auth-channels-file",
    default=DEFAULT_CONFIG_CHANNELS,
    help=(
        "read channel names from this file; you can also set the list of channels via "
        "the TURBOT_CHANNELS environment variable separated by ; (semicolon)."
    ),
)
@click.option(
    "--db-dir", default=DEFAULT_DB_DIR, help="use this directory for user db files"
)
@click.version_option(version=__version__)
@click.option(
    "--dev",
    default=False,
    is_flag=True,
    help="Development mode, automatically reload bot when source changes",
)
def main(
    log_level, verbose, bot_token_file, channel, auth_channels_file, db_dir, dev,
):  # pragma: no cover
    auth_channels = get_channels(auth_channels_file) + list(channel)
    if not auth_channels:
        print("error: you must provide at least one authorized channel", file=sys.stderr)
        sys.exit(1)

    # ensure transient application directories exist
    db_dir.mkdir(exist_ok=True)
    TMP_DIR.mkdir(exist_ok=True)

    client = Turbot(
        token=get_token(bot_token_file),
        channels=auth_channels,
        db_dir=db_dir,
        log_level=getattr(logging, "DEBUG" if verbose else log_level),
    )

    apply_migrations()

    if dev:
        reloader = hupper.start_reloader("turbot.main")
        reloader.watch_files(client.assets.files())

    client.run()


if __name__ == "__main__":
    main()



import re
from os.path import dirname, realpath
from pathlib import Path
from string import Template

import pandas as pd
from unidecode import unidecode
from yaml import load

try:
    from yaml import CLoader as Loader
except ImportError:  # pragma: no cover
    from yaml import Loader

PACKAGE_ROOT = Path(dirname(realpath(__file__)))
DATA_DIR = PACKAGE_ROOT / "data"
STRINGS_DATA_FILE = DATA_DIR / "strings.yaml"


def load_strings():
    with open(STRINGS_DATA_FILE) as f:
        return load(f, Loader=Loader)


class Asset:
    def __init__(self, info):
        self.info = info

    @property
    def all(self):
        return self.info["items"]

    @property
    def data(self):
        return self.info["data"]


class Assets:
    def __init__(self):
        self.collectables = ["fossils", "bugs", "fish", "art", "songs"]
        all_items = set()
        self.asset = {}
        for kind in self.collectables:
            file = DATA_DIR / f"{kind}.csv"
            data = pd.read_csv(file)
            items = frozenset(data.drop_duplicates(subset="name").name.tolist())
            all_items.update(items)
            self.asset[kind] = {"file": file, "data": data, "items": items}
        self.all_collectable_items = frozenset(all_items)

    def validate(self, items, *, kinds=None):
        def clean(item):
            return re.sub(" ?k.k. ?", "", unidecode(item).lower())

        items = [clean(item) for item in items]
        valid = {}
        for kind in self.collectables:
            if kinds and isinstance(kinds, list) and kind not in kinds:
                continue
            valid_items = set()
            for item in self.asset[kind]["items"]:
                cleaned_item = clean(item)
                if cleaned_item in items:
                    valid_items.add(item)
                    items.remove(cleaned_item)
            valid[kind] = valid_items
        return valid, items

    def files(self):  # pragma: no cover
        return [STRINGS_DATA_FILE] + [
            self.asset[kind]["file"] for kind in self.collectables
        ]

    def __getitem__(self, kind):
        return Asset(self.asset[kind])


def s(key, **kwargs):
    """Returns a string from data/strings.yaml with subsitutions."""
    if not s.strings:
        s.strings = load_strings()

    data = s.strings.get(key, "")
    assert data, f"error: missing strings key: {key}"
    return Template(data).substitute(kwargs)


s.strings = None



from io import StringIO
from pathlib import Path

import pandas as pd


class StoreFrame(pd.DataFrame):
    _metadata = ["data_type"]

    @property
    def _constructor(self):
        s = StoreFrame
        s.data_type = self.data_type
        return s

    @property
    def _constructor_sliced(self):
        s = pd.Series
        s.data_type = self.data_type
        return s


class Data:
    """Persistent and in-memory store for user data."""

    def __init__(self, *, db_dir):
        self.db_dir = db_dir
        self.config = {
            "fossils": {"columns": {"author": "int64", "name": "str"},},
            "art": {"columns": {"author": "int64", "name": "str"}},
            "fish": {"columns": {"author": "int64", "name": "str"}},
            "bugs": {"columns": {"author": "int64", "name": "str"}},
            "songs": {"columns": {"author": "int64", "name": "str"}},
            "prices": {
                "columns": {
                    "author": "int64",
                    "kind": "object",
                    "price": "int64",
                    "timestamp": "datetime64[ns, UTC]",
                },
            },
            "users": {
                "columns": {
                    "author": "int64",
                    "hemisphere": "str",
                    "timezone": "str",
                    "island": "str",
                    "friend": "str",
                    "fruit": "str",
                    "nickname": "str",
                    "creator": "str",
                },
            },
        }
        for data_type in self.config.keys():
            self.config[data_type]["file"] = db_dir / f"{data_type}.csv"
        self.in_memory = {data_type: None for data_type in self.config.keys()}

    def file(self, data_type):
        return self.config[data_type]["file"]

    def commit(self, sf):
        f_out = self.config[sf.data_type]["file"]
        sf.to_csv(f_out, index=False)
        self.in_memory[sf.data_type] = sf

    def __getattr__(self, attr):
        if attr not in self.config.keys():
            raise RuntimeError(f"there is no data store for {attr}")

        cfg = self.config[attr]
        columns = cfg["columns"]

        def add_metadata(df):
            sf = StoreFrame(df.fillna("").astype(columns))
            sf.data_type = attr
            return sf

        df = self.in_memory[attr]
        if df is not None:
            return add_metadata(df)

        f_in = cfg["file"]
        cnames = list(columns.keys())
        dtypes = list(columns.values())
        if Path(f_in).exists():
            parse_dates = any("datetime" in dtype for dtype in dtypes)
            df = pd.read_csv(f_in, names=cnames, parse_dates=parse_dates, skiprows=1)
        else:
            df = pd.read_csv(StringIO(""), names=cnames, dtype=columns)

        return add_metadata(df)
